"""
    ExtrapolationControlPanel.py

    Copyright (c) 2018-2025, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy.KekLib.OurTtk import ttk, TtkScale
from molass_legacy.KekLib.TkSupplements import BlinkingFrame
from ExtrapolationAnimation import ExtrapolationAnimationDialog, TOO_SMALL_TO_PLOT
from molass_legacy.SerialAnalyzer.DevSettings import get_dev_setting
from molass_legacy.KekLib.TkCustomWidgets import FileEntry
from molass_legacy.SerialAnalyzer.SerialDataUtils import serial_np_loadtxt

SCALE_MAX       = 0.2
SCALE_DIGITS    = 2
ENABLE_ANIMATE  = False
ENABLE_KNOWN_INPUT = False

class ControlPanel( Tk.Frame ):
    def __init__(self, parent, dialog, pno, ad, paired_range, row, selected, weights, ignore_bq, editor=None):
        Tk.Frame.__init__( self, parent )
        self.known = Tk.IntVar()

        if ENABLE_KNOWN_INPUT:
            cb= Tk.Checkbutton(self, text="Known", variable=self.known)
            cb.pack(anchor=Tk.W)

        self.unknown_panel = UnknownControlPanel(self, dialog, pno, ad, paired_range, row, selected, weights, ignore_bq, editor)
        self.unknown_panel.pack(anchor=Tk.W, padx=0)

        if ENABLE_KNOWN_INPUT:
            self.known_panel = KnownControlPanel(self, dialog, pno, ad, row, selected, self.unknown_panel)
            self.known_panel.pack(anchor=Tk.W, fill=Tk.X, expand=1)
            self.known_panel.pack_forget()

            self.known.trace("w", self.known_tracer)

    def known_tracer(self, *args):
        known = self.known.get()
        if known:
            self.unknown_panel.pack_forget()
            self.known_panel.pack(anchor=Tk.W)
        else:
            self.known_panel.pack_forget()
            self.unknown_panel.pack(anchor=Tk.W)

class UnknownControlPanel( Tk.Frame ):
    def __init__(self, parent, dialog, pno, ad, paired_range, row, selected, weights, ignore_bq, editor):
        Tk.Frame.__init__(self, parent)

        self.dialog = dialog
        self.editor = editor
        self.pno = pno
        self.ad = ad
        self.paired_range = paired_range
        self.row = row
        self.selected = selected
        # print( 'ControlPanel', [pno, ad, paired_range], selected )
        self.weights = weights
        self.solving_required = False
        self.add_conc_const = get_dev_setting('add_conc_const')
        self.individual_bq_ingore = get_dev_setting('individual_bq_ingore')

        scale_frame = Tk.Frame( self )
        scale_frame.pack()
        self.button_frame = BlinkingFrame( self )
        self.button_frame.pack()

        grid_row = 0
        base_drift_label = Tk.Label( scale_frame, text="Baseline Drift Penality" )
        base_drift_label.grid( row=grid_row, column=0, sticky=Tk.W + Tk.S )

        self.base_drift_scale = TtkScale( scale_frame, from_=0, to=SCALE_MAX,
                                        # resolution=0.1, sliderlength=10,
                                        length=160, orient=Tk.HORIZONTAL,
                                        showvalue=True, showvalue_y=14, digits=SCALE_DIGITS,
                                        style='my.Horizontal.TScale', slim=True )
        self.base_drift_scale.grid( row=grid_row, column=1 )
        self.base_drift_scale.set( self.weights[1] )

        if self.add_conc_const:
            self.base_drift_scale.variable.trace('w', self.parameter_tracer)
        else:
            base_drift_label.grid_forget()
            self.base_drift_scale.grid_forget()

        grid_row += 1
        a_smoothness_label = Tk.Label( scale_frame, text="A(q) Smoothness Penality" )
        a_smoothness_label.grid( row=grid_row, column=0, sticky=Tk.W + Tk.S )

        self.a_smoothness_scale = TtkScale( scale_frame, from_=0, to=SCALE_MAX,
                                        # resolution=0.1, sliderlength=10,
                                        length=160, orient=Tk.HORIZONTAL,
                                        showvalue=True, showvalue_y=14, digits=SCALE_DIGITS,
                                        style='my.Horizontal.TScale', slim=True )
        self.a_smoothness_scale.grid( row=grid_row, column=1 )
        self.a_smoothness_scale.set( self.weights[2] )

        grid_row += 1
        self.b_smoothness_label = Tk.Label( scale_frame, text="B(q) Smoothness Penality" )
        self.b_smoothness_row = grid_row

        self.b_smoothness_scale = TtkScale( scale_frame, from_=0, to=SCALE_MAX,
                                        # resolution=0.1, sliderlength=10,
                                        length=160, orient=Tk.HORIZONTAL,
                                        showvalue=True, showvalue_y=14, digits=SCALE_DIGITS,
                                        style='my.Horizontal.TScale', slim=True )
        self.b_smoothness_scale.set( self.weights[3] )

        self.ignore_bq = Tk.IntVar()
        self.ignore_bq.set(ignore_bq)
        self.bq_widgets_update()

        if dialog.aq_smoothness:
            self.a_smoothness_scale.variable.trace('w', self.parameter_tracer)
            self.b_smoothness_scale.variable.trace('w', self.parameter_tracer)
        else:
            a_smoothness_label.grid_forget()
            self.a_smoothness_scale.grid_forget()
            self.b_smoothness_label.grid_forget()
            self.b_smoothness_scale.grid_forget()

        grid_row += 1
        a_neg_label = Tk.Label( scale_frame, text="A(q) Positivity Penality" )
        a_neg_label.grid( row=grid_row, column=0, sticky=Tk.W + Tk.S )

        style = ttk.Style(self)
        style.configure('my.Horizontal.TScale', sliderlength=10)

        self.a_neg_scale = TtkScale( scale_frame, from_=0, to=SCALE_MAX,
                                        # resolution=0.1, sliderlength=10,
                                        length=160, orient=Tk.HORIZONTAL,
                                        showvalue=True, showvalue_y=14, digits=SCALE_DIGITS,
                                        style='my.Horizontal.TScale', slim=True )
        self.a_neg_scale.grid( row=grid_row, column=1 )
        self.a_neg_scale.set( self.weights[0] )

        if dialog.aq_positivity:
            self.a_neg_scale.variable.trace('w', self.parameter_tracer)
        else:
            a_neg_label.grid_forget()
            self.a_neg_scale.grid_forget()

        if self.individual_bq_ingore:
            ignore_bq_cb = Tk.Checkbutton( self.button_frame, text="Ignore B(q)", variable=self.ignore_bq, state=Tk.NORMAL )
            ignore_bq_cb.pack( side=Tk.LEFT, padx=5, pady=5 )
            self.ignore_bq.trace('w', self.ignore_bq_tracer)

        self.button_frame.objects = []

        if ENABLE_KNOWN_INPUT:
            solve_button = Tk.Button( self.button_frame, text="Solve", command=self.solve_plot )
            solve_button.pack(padx=5, pady=5)
            self.button_frame.objects.append(solve_button)

        if ENABLE_ANIMATE:
            amin_button = Tk.Button( self.button_frame, text="Animate", command=self.make_animation )
            amin_button.pack(padx=5, pady=5)
            self.button_frame.objects.append(amin_button)

        range_button = Tk.Button( self.button_frame, text="Range", command=self.show_range_inspector )
        range_button.pack(padx=5, pady=5 )

        rank_button = Tk.Button( self.button_frame, text="Rank", command=self.show_cdi_dialog )
        rank_button.pack(padx=5, pady=5 )

        denss_button = Tk.Button( self.button_frame, text="DENSS", command=self.show_denss_dialog )
        denss_button.pack(padx=5, pady=5 )

    def set_C_matrix( self, C ):
        self.C = C

    def bq_widgets_update( self ):
        if self.dialog.aq_smoothness:
            ignore_bq = self.ignore_bq.get()
            if ignore_bq:
                self.b_smoothness_label.grid_forget()
                self.b_smoothness_scale.grid_forget()
            else:
                self.b_smoothness_label.grid( row=self.b_smoothness_row, column=0, sticky=Tk.W + Tk.S )
                self.b_smoothness_scale.grid( row=self.b_smoothness_row, column=1 )

    def ignore_bq_tracer( self, *args ):
        if self.solving_required:
            self.button_frame.stop()
            self.solving_required = False
        else:
            self.button_frame.start()
            self.solving_required = True
        self.bq_widgets_update()

    def parameter_tracer( self, *args ):
        self.solving_required = True
        self.button_frame.start()

    def get_penalty_weights( self ):
        # always return all weights
        return [    self.a_neg_scale.get(),
                    self.base_drift_scale.get(),
                    self.a_smoothness_scale.get(),
                    self.b_smoothness_scale.get() ]

    def solve_plot( self ):
        self.button_frame.stop()
        self.solving_required = False
        self.dialog.config(cursor='wait')
        self.dialog.update()
        penalty_weights = self.get_penalty_weights()
        self.dialog.solve_plot( self.row, self.selected, penalty_weights=penalty_weights, ignore_bq=self.ignore_bq.get() )
        self.dialog.canvas_draw()
        self.bq_widgets_update()
        self.dialog.config(cursor='')
        self.dialog.update()

    def make_animation( self ):
        self.button_frame.stop()
        self.solving_required = False
        self.dialog.config(cursor='wait')
        self.dialog.update()
        penalty_weights = self.get_penalty_weights()
        anim_data = self.dialog.solve_plot( self.row, self.selected, penalty_weights=penalty_weights, ignore_bq=self.ignore_bq.get(), animation=True )
        self.dialog.canvas_draw()
        self.dialog.config(cursor='')
        self.dialog.update()

        # ridge = self.dialog.data[]
        # print( 'cnv_ranges=', self.dialog.cnv_ranges[self.row] )
        top_x = self.paired_range.top_x
        ridge = self.dialog.data[:,top_x] / self.dialog.mc_vector[top_x]
        ridge[ridge < TOO_SMALL_TO_PLOT] = 0    # to avoid these values from deforming the figure
        dialog = ExtrapolationAnimationDialog( self.dialog, self.row, self.dialog.q, ridge, anim_data, penalty_weights, ignore_bq=self.ignore_bq.get() )
        dialog.show()

    def show_range_inspector( self ):
        from .RangeInspector import RangeInspectorDialog
        dialog = self.dialog
        try:
            solver = dialog.solver.impl
        except:
            solver = dialog.solver
        init_result = dialog.solver_results[self.row]
        conc_depend = dialog.popts.conc_depend
        inspector = RangeInspectorDialog(dialog.parent, solver, dialog.j0, self.paired_range, self.selected, self.ad,
                    conc_depend=conc_depend,
                    init_result=init_result)
        ret = inspector.show()
        if ret:
            new_range = inspector.get_range(shifted=False)
            self.aplly_new_range(*new_range)
            new_shifted_range = inspector.get_range(shifted=True)
            self.editor.update_range(self.pno, self.ad, *new_shifted_range)
            self.solve_plot()

    def show_cdi_dialog(self):
        from molass_legacy.Conc.CdInspection import CdInspectionDailog
        print( self.pno, self.ad, self.paired_range )
        print( self.dialog.q.shape, self.dialog.data.shape )

        q = self.dialog.q

        f, t = self.paired_range[self.ad]
        eslice = slice(f, t+1)
        M = self.dialog.data[:, eslice]
        E = self.dialog.error[:, eslice]
        from_ax = self.dialog.axis_array[self.row][0]
        xray_scale = self.dialog.sd.get_xray_scale()
        dialog = CdInspectionDailog(self.dialog.parent, self.dialog, M, E, self.C, q, eslice, from_ax=from_ax, xray_scale=xray_scale)
        dialog.show()

    def aplly_new_range(self, f, t):
        print('aplly_new_range', (f, t))
        self.paired_range.update_range(self.ad, f, t)
        self.dialog.update_to_solve_range(self.row, f, t)
        """
        TODO:
            refactoring: unifiy these range infos
        """

    def get_file_name(self):
        print('show_denss_dialog', self.pno, self.ad, self.paired_range, self.row)
        if len(self.paired_range.get_fromto_list()) == 1:
            ad_str = 'bth'
        else:
            ad_str = 'asc' if self.ad == 0 else 'dsc'
        file_name = 'pk%d_%s_A.dat' % (self.pno+1, ad_str)
        return file_name

    def show_denss_dialog( self ):
        from molass_legacy._MOLASS.Version import is_developing_version
        if is_developing_version():
            from importlib import reload
            import molass_legacy.DENSS.DenssGui
            reload(molass_legacy.DENSS.DenssGui)
        from molass_legacy.DENSS.DenssGui import DenssGuiDialog
        A, B, Z, E, _, C = self.dialog.solver_results[self.row]
        dialog = DenssGuiDialog( self.dialog.parent, self.dialog.q, A, E[0], self.get_file_name() )
        dialog.show()

class KnownControlPanel( Tk.Frame ):
    def __init__(self, parent, dialog, pno, ad, row, selected, unknown_panel):
        width = unknown_panel.cget('width')
        height = unknown_panel.cget('height')
        Tk.Frame.__init__( self, parent, width=width, height=height)

        self.dialog = dialog
        self.row = row
        self.selected = selected

        label = Tk.Label(self, text="Enter a file path for your known scattering curve.")
        label.pack(anchor=Tk.W)

        self.known_data_file = Tk.StringVar()
        entry = FileEntry(self, textvariable=self.known_data_file, width=45, on_entry_cb=self.on_entry_file)
        entry.pack()

        num_spaces = 3      # actually variable
        for k in range(num_spaces):
            space = Tk.Label(self)
            space.pack()

    def on_entry_file(self):
        file_path = self.known_data_file.get()
        self.data, _ = serial_np_loadtxt(file_path)
        self.dialog.plot_known(self.row, self.selected, self.data)
        self.dialog.su_btn_blink.start()



# coding: utf-8
"""

    ReadOnlyText.py

    copied from
    Readonly tkinter text widget
    https://stackoverflow.com/questions/21873195/readonly-tkinter-text-widget/21882093

"""
import tkinter as Tk

# This is the list of all default command in the "Text" tag that modify the text
commandsToRemove = (
"<Control-Key-h>",
"<Meta-Key-Delete>",
"<Meta-Key-BackSpace>",
"<Meta-Key-d>",
"<Meta-Key-b>",
"<<Redo>>",
"<<Undo>>",
"<Control-Key-t>",
"<Control-Key-o>",
"<Control-Key-k>",
"<Control-Key-d>",
"<Key>",
"<Key-Insert>",
"<<PasteSelection>>",
"<<Clear>>",
"<<Paste>>",
"<<Cut>>",
"<Key-BackSpace>",
"<Key-Delete>",
"<Key-Return>",
"<Control-Key-i>",
"<Key-Tab>",
"<Shift-Key-Tab>"
)

class ReadOnlyText(Tk.Text):
    tagInit = False

    def init_tag(self):
        """
        Just go through all binding for the Text widget.
        If the command is allowed, recopy it in the ROText binding table.
        """
        for key in self.bind_class("Text"):
            if key not in commandsToRemove:
                command = self.bind_class("Text", key)
                self.bind_class("ReadOnlyText", key, command)
        ReadOnlyText.tagInit = True

    def __init__(self, *args, **kwords):
        Tk.Text.__init__(self, *args, **kwords)
        if not ReadOnlyText.tagInit:
            self.init_tag()

        # Create a new binding table list, replace the default Text binding table by the ROText one
        bindTags = tuple(tag if tag!="Text" else "ReadOnlyText" for tag in self.bindtags())
        self.bindtags(bindTags)

class CopyableLabel(ReadOnlyText):
    def __init__( self, parent, text, *args, **kwargs ):
        ReadOnlyText.__init__( self, parent, *args, **kwargs )
        self.insert( Tk.END, text )
        lable = Tk.Label( parent )
        bg = lable.cget( 'bg' )
        self.config( relief=Tk.FLAT, height=1, width=len(text), bg=bg )

    def config( self, text=None, *args, **kwargs ):
        if text is not None:
            self.delete( '1.0', Tk.END )
            self.insert( Tk.END, text )
            self.config( width=len(text) )
        ReadOnlyText.config( self, *args, **kwargs )



# coding: utf-8
"""
    KekTools.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""

import sys
import traceback
import time
import numpy                as np
import scipy.stats          as stats
import matplotlib.pyplot    as plt
from bisect                 import bisect_right
from IntensityData          import IntensityData, ACCEPTABLE_BASIC_QUALITY
from ErrorResult            import ErrorResult

DEBUG = False
MINIMUM_INTERVAL_SIZE   = 5
RVALUE_ALLOWANCE        = 0.01

class DummyFit:
    def __init__( self, **entries ): 
        self.__dict__.update(entries)

class AutorgKek:
    def __init__( self, file=None ):

        if file is None: return

        if type( file ) == str or type( file ) == np.ndarray:
            try:
                intensity = IntensityData( file, add_smoother=False )
            except ( ValueError, IndexError ):
                # these errors are known for very low quality data
                # which include many negative values.
                intensity = None
        elif type( file ) == IntensityData:
            intensity = file
        else:
            assert( False )

        self.intensity  = intensity
        self.interval   = None
        self.guinier    = None
        self.optimal_interval_q = None

    def run( self, robust=False, optimize=True,
                qrg_limits=[ 0.0, 1.3 ],
                retry_mode=False
           ):

        self.fit = DummyFit()

        self.X  = X = self.intensity.X
        self.Y  = Y = self.intensity.Y
        self.smoother   = self.intensity.smoother

        self.m   = m = np.argmax( Y )
        print( 'm=', m )

        self.small_boundary = small_boundary = int( len( Y ) ) - 1
        stop_j  = self.seartch_for_stop( m, self.small_boundary )
        start_i = self.seartch_for_start( m, stop_j )
        stop_j  = self.seartch_for_stop( start_i, stop_j )

        slope, intercept, r_value, p_value, std_err = stats.linregress( X[start_i:stop_j], Y[start_i:stop_j] )
        Q   = np.sqrt( self.X[stop_j-1] )
        Rg  = np.sqrt( 3*(-slope) ) if slope < 0 else None
        print( (start_i, stop_j), 'Rg=', Rg, 'Q*Rg=', Q*Rg )

        # self.solve_by_annealing()

        self.result = ErrorResult()
        return self.result

    def seartch_for_stop( self, start, stop ):
        r_list = []
        s_list = []
        for j in range( start + MINIMUM_INTERVAL_SIZE, stop ):
            slope, intercept, r_value, p_value, std_err = stats.linregress( self.X[start:j], self.Y[start:j] )
            r_list.append( r_value )
            s_list.append( std_err )
            Q   = np.sqrt( self.X[j] )
            if slope < 0:
                Rg  = np.sqrt( 3*(-slope) )
                if Q*Rg > 1.3:
                    break
            else:
                continue

        r_array = np.array( r_list )
        s_array = np.array( s_list )
        rm = np.argmin( r_array )
        print( 'rm=', rm, self.X[rm], 'Rg=', Rg )
        stop_j = min( j, rm + 1)

        # self.debug_plot( r_array, s_array, start )    # broken

        return stop_j

    def seartch_for_start( self, start, stop ):
        r_list = []
        s_list = []
        # qrg_list = []
        Q   = np.sqrt( self.X[stop-1] )

        for i in range( start, stop - MINIMUM_INTERVAL_SIZE ):
            slope, intercept, r_value, p_value, std_err = stats.linregress( self.X[i:stop], self.Y[i:stop] )
            r_list.append( r_value )
            s_list.append( std_err )
            Rg  = np.sqrt( 3*(-slope) ) if slope < 0 else None
            # qrg_list.append( Q*Rg )

        r_array = np.array( r_list )
        s_array = np.array( s_list )
        rm = np.argmin( r_array )
        print( 'rm=', rm, self.X[rm], 'Rg=', Rg )
        return rm

    def solve_by_annealing( self ):
        from SimulatedAnnealing     import SimulatedAnnealing

        X   = self.X
        Y   = self.Y
        m   = self.m
        small_boundary  = self.small_boundary

        def xconstaints( x ):
            return x[0] + 5 < x[1]

        def f( x ):
            i = int( x[0] )
            j = int( x[1] )
            slope, intercept, r_value, p_value, std_err = stats.linregress( X[i:j], Y[i:j] )
            return r_value + std_err

        anneal = SimulatedAnnealing()
        anneal.minimize( f, xrange=[ [m, small_boundary], [m, small_boundary] ], start=[m, small_boundary], seed=1234,
                        xconstaints=xconstaints )

        print( 'anneal.start=', anneal.start )
        print('Best solution: ' + str(anneal.xc))
        print('Best objective: ' + str(anneal.fc))

    def debug_plot( self, r_array, s_array, start_i, stop_j ):
        fig = plt.figure( figsize=( 18, 12) )
        ax1  = fig.add_subplot( 331 )
        ax2  = fig.add_subplot( 332 )
        ax3  = fig.add_subplot( 333 )
        ax5  = fig.add_subplot( 335 )
        ax8  = fig.add_subplot( 338 )
        ax1.plot( self.intensity.X, self.intensity.Y, 'o', markersize=3 )
        xmin1, xmax1 = ax1.get_xlim()
        ymin1, ymax1 = ax1.get_ylim()

        xmin2   = xmin1 * ( 257/256 ) + xmax1 * ( -1/256 )
        xmax2   = xmin1 * 61/64 + xmax1 * 3/64
        ymin2   = ymin1 * 3/8 + ymax1 * 5/8
        ymax2   = ymax1
        ax2.set_xlim( xmin2, xmax2 )
        ax2.set_ylim( ymin2, ymax2 )
        ax2.plot( self.intensity.X, self.intensity.Y, 'o', markersize=3 )

        xmin3   = xmin1 * ( 257/256 ) + xmax1 * ( -1/256 )
        xmax3   = xmin1 * 125/128 + xmax1 * 3/128
        ymin3   = ymin1 * 2/16 + ymax1 * 14/16
        ymax3   = ymax1
        ax3.set_xlim( xmin3, xmax3 )
        ax3.set_ylim( ymin3, ymax3 )
        ax3.plot( self.intensity.X, self.intensity.Y, 'o', markersize=3 )

        ax5.set_xlim( xmin2, xmax2 )
        ax5.set_ylim( -1, -0.95 )
        ax5.plot( self.X[start_i:stop_j], r_array )
        ax8.set_xlim( xmin2, xmax2 )
        ax8.set_ylim( 0, 15 )
        ax8.plot( self.X[start_i:stop_j], s_array )

        for ax in [ ax1, ax2, ax3 ]:
            ax.plot( self.intensity.X[self.m], self.intensity.Y[self.m], 'o', color='red' )

        fig.tight_layout()
        plt.show()

def autorg( file, robust=False, optimize=True,
            qrg_limits=[ 0, 1.3 ], qrg_limits_apply=True ):
    autorg_ = AutorgKek( file  )
    autorg_.run( robust=robust, optimize=optimize,
                    qrg_limits=qrg_limits )
    return autorg_.result, autorg_.intensity



"""
    HardSphereDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt

def demo1():
    from importlib import reload
    import SimTools.HardSphere
    reload(SimTools.HardSphere)
    from SimTools.HardSphere import get_model_data
    import SimTools.EoiiPlotUtils
    reload(SimTools.EoiiPlotUtils)
    from SimTools.EoiiPlotUtils import plot_eoii

    qv = np.linspace(0.005, 0.4, 200)
    jv = np.arange(300)
    rg = 35
    h = 1
    mu = 150
    sigma = 20
    tau = 0
    K = 0.5
    D, P, C = get_model_data(qv, jv, rg, h, mu, sigma, tau, K=K)

    title = "Effect of Interparticle Interactions Simulation with K=%g" % K

    De, Pe, Ce = get_model_data(qv, jv, rg, h, mu, sigma, tau, K=K, error=0.008)

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(18,10))
        fig.suptitle(title, fontsize=20)
        ax1, ax2, ax3 = axes[0,:]
        ax1.set_title("Elution View", fontsize=16)
        ax2.set_title("Scattering View (Linear)", fontsize=16)
        ax3.set_title("Scattering View (Log)", fontsize=16)
        plot_eoii(qv, jv, D, P, C, axes[0,:])

        ax1, ax2, ax3 = axes[1,:]
        ax1.set_title("Noisy Data Elution View", fontsize=16)
        ax2.set_title("Noisy Data Scattering View (Linear)", fontsize=16)
        ax3.set_title("Noisy Data Scattering View (Log)", fontsize=16)
        plot_eoii(qv, jv, De, Pe, Ce, axes[1,:])

        fig.tight_layout()
        plt.show()

def demo():
    from importlib import reload
    import SimTools.HardSphere
    reload(SimTools.HardSphere)
    from SimTools.HardSphere import get_model_data
    import SimTools.EoiiPlotUtils
    reload(SimTools.EoiiPlotUtils)
    from SimTools.EoiiPlotUtils import plot_eoii

    qv = np.linspace(0.005, 0.4, 200)
    jv = np.arange(300)
    rg = 35
    h = 1
    mu = 150
    sigma = 20
    tau = 0
    K = 0.5
    D, P, C = get_model_data(qv, jv, rg, h, mu, sigma, tau, K=K)

    title = "Eoii Algorithm using Aparent Data"

    i = bisect_right(qv, 0.02)

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(18,10))
        fig.suptitle(title, fontsize=20)
        ax1, ax2, ax3 = axes[0,:]
        ax1.set_title("Elution View", fontsize=16)
        ax2.set_title("Scattering View (Linear)", fontsize=16)
        ax3.set_title("Scattering View (Log)", fontsize=16)
        plot_eoii(qv, jv, D, P, C, axes[0,:])

        A = D.copy()
        cy_init = A[i,:]
        cy = cy_init
        cy2 = cy**2
        C_ = np.array([cy, cy2])
        C_inv = np.linalg.pinv(C_)
        P_ = D @ C_inv
        aq, bq = P_.T
        ax4 = axes[1,0]
        ax4.plot(qv, aq)
        ax4.plot(qv, bq)
        B = bq[:,np.newaxis] @ cy2[np.newaxis,:]
        A_ = D - B

        ax5 = axes[1,1]
        ax5.plot(jv, cy_init)
        ax5.plot(jv, A_[i,:], ":")

        fig.tight_layout()
        plt.show()



"""
    SimpleGuinier.py

    Copyright (c) 2017-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.interpolate import LSQUnivariateSpline
from scipy import stats
import logging
from bisect import bisect_right
from molass_legacy.KekLib.NumpyUtils import np_loadtxt_robust
from molass_legacy.KekLib.OurCurvature import curvature_curve
from SimpleGuinierScore import ( MIN_GUINIER_SIZE, ZERO_SCORE_VECTOR, END_GUINIER_SIZE,
                                    QUALITY_BOUNDARY_HIGH, USE_GUINIER_PLOT_LENGTH,
                                    compute_rg,
                                    compute_expected_size,
                                    evaluate_rg_consistency,
                                    evaluate_guinier_interval,
                                    compute_end_consistency,
                                    compute_fwd_consistency )
from molass_legacy._MOLASS.SerialSettings import get_setting

NUM_INITIAL_IN_KNOTS    = 2
KNOT_NO_ADDITION        = 0
KNOT_DROPPING_HEAD      = 1
KNOT_SOARING_HEAD       = 2
MIN_NUM_HEAD_DROPS      = 10
FWD_NUM_HEAD_DROPS      = 20            # required for 20170301-207
TODO_DROPPING_RATIO     = 1.5
TODO_SOARING_RATIO      = 0.3           # required < 0.54 for SUB_TRN1PY-189, < 0.40 for TSsonme2-224
GOOD_SOARING_RATIO      = 1.8           # required > 1.64 for SUB_TRN1PY-195, < 1.98 for 20160227-0
ADEDUITE_R_VALUE        = -0.998
CANDIDATE_R_VALUE       = -0.96
ALTERNATIVE_R_VALUE     = -0.95
MONOTONICITY_SCALE      = 0.5
QUALITY_BOUNDARY_WORST  = 0.005         # 0.0073 for Matsumura2-172
QUALITY_BOUNDARY_LOW    = 0.02          # for SUB_TRN1_2-206, < 0.022 for 20161202-168
QUALITY_BOUNDARY_QRG    = 0.5           # 
QUALITY_BOUNDARY_BEST   = 0.95
ALLOW_FWD_CONSISTENCY   = 0.5
ALLOW_FWD_CONSISTENCY2  = 0.1
BACK_R_VALUE_PENALTY    = 0.3
MAX_NUM_GOING_BACK      = 40
PEAK_FIND_ROTATION      = -np.pi/6      # -π/6, i.e. 30°clockwise
PEAK_FIND_R_VALUE       = -0.03         # required for AhRR-10
PEAK_FIND_R_VALUE_ALLOW = -0.05
PEAK_FIND_TOO_LONG      = 40            # < 44 for SUB_TRN1-165, < 70 for 3/9-196
PEAK_FIND_RETRIABLE_LEN = 25
PEAK_FIND_PERCENTILE    =  5
PEAK_FIND_QUALITY_BNDRY = 50            # 58.5 for AhRR-12
PEAK_FIND_MIN_RG_RATIO  = 3
PEAK_FIND_RG_RATIO_LOW  = 2             # < 2.16 for TRN1PY-195, > 1.38 for Matsumura-258,  > 1.06 for 12/2-113
PEAK_FIND_MIN_AR_RATIO  = 0.2           # > 0.18 for Open01-232
PEAK_FIND_QRG_LIMIT     = 1.5           # > 
PEAK_FIND_MIN_CURVATURE = 0.0003        # < 0.00034 for SUB_TRNPY-193, > 0.00018 for Close01-205
MONOTONICITY_FWD_OK     = 0.5           # > 0.30 for Matsumura-258
TANGENT_DIFF_BOUNDARY   = 0.006         # > 0.005 for Open01-257
GOOD_MONOTONICITY_RATIO = 3             # > 2.15 for AhRR-10
TRY_PEAK_MIN_OFFSET     = 3
UP_ADJUST_X_DEV_BNDRY   = 0.01          # > 0.0032 for SUB_TRN1-160, < 0.017 for Matsumura2-172
LINREG_INDEX            = slice(0,3,2)
LINREG_STDERR           = 4
MIN_BETTER_RG_RATIO     = 75            # > 69 for 20160227-0,  < 81 for AHRR-12,  < 131 for AhRR-133
MAX_START_Q2            = 0.0005        # 
MIN_START_FROM_PEAK     = 5             # required for 20160628/AIMdim01_00179_sub.dat
SMALL_Q2_BOUNDARY       = MAX_START_Q2/2
SCORE_RATIO_BOUNDARY    = 0.7
WST_AT_LEAST_TRIALS     = 2
LQ_AT_LEAST_TRIALS      = 3

EXCEPTION_LOG   = True
DEBUG_PLOT      = False

def rotate( th, wx, wy ):
    cx  = wx[-1]
    cy  = wy[-1]
    wx_ = wx - cx
    wy_ = wy - cy
    c   = np.cos( th )
    s   = np.sin( th )
    return cx + ( wx_*c - wy_* s ), cy + ( wx_*s + wy_* c )

def compute_monotonicity( y ):
    height = y[0] - y[-1]

    if height > 0:
        gry     = np.gradient( y )
        ag_sum  = np.sum( np.abs( gry ) )
        ag_ratio = ag_sum / height
        monotonicity = np.exp( MONOTONICITY_SCALE *( 1 - ag_ratio ) )
        avg_gradient = np.average( gry )
    else:
        monotonicity = 0
        avg_gradient = 0

    return monotonicity, avg_gradient

class SimpleGuinier:
    def __init__( self, file, curvature=False, basic_quality_only=False, anim_data=False, debug_plot=False ):
        self.logger = logging.getLogger( __name__ )
        self.debug_plot = debug_plot or DEBUG_PLOT
        if type( file ) == str:
            self.data, _ = np_loadtxt_robust( file )
        elif type( file ) == np.ndarray:
            self.data = file
        else:
            raise RuntimeError()

        # print( self.data.shape )
        self.curvature  = curvature
        self.anim_data  = anim_data
        if self.anim_data:
            self.best_history_dict = {}
        self.x_ = x_ = self.data[ :, 0 ]
        self.y_ = y_ = self.data[ :, 1 ]
        self.e_ = e_ = self.data[ :, 2 ]
        positive = np.logical_and( y_ > 0, e_ > 0 )

        self.qrg_stop = None
        self.rg_ratio_for_better_peak = None
        self.head_gap_ratio = None
        self.basic_quality = None
        self.too_long_peak_candidate = None
        self.anim_guinier_stop = None
        self.peak_type  = None
        self.try_ret    = None
        self.tester_log_writer = None

        try:
            self.x = x_[positive]
            self.y = y_[positive]
            self.e = e_[positive]
            self.dx = x_[1] - x_[0]
            self.x2 = self.x**2
            self.log_y = np.log(self.y)
            self.log_y_err = 1/self.y * self.e  # error propagation of logarithm: error log(y) = 1/y * y_error
            if len( self.log_y ) > MIN_GUINIER_SIZE * 4:
                self.make_smooth_line()
                self.evaluate_basic_quality()
            else:
                # as in 20170426-21
                self.logger.warning("SimpleGuinier aborted due to %d <= %d", len( self.log_y ), MIN_GUINIER_SIZE * 4)
        except:
            if EXCEPTION_LOG:
                from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
                etb = ExceptionTracebacker()
                self.logger.error( etb )
            else:
                pass

        if basic_quality_only:
            return

        if self.basic_quality is None:
            self.set_guinier_null_result()
        else:
            self.guinier_interval()

    def make_smooth_line(self, debug=False):
        n = len(self.x)
        points = [ n ]
        for i in range(NUM_INITIAL_IN_KNOTS):
            n //= 2
            points.append(n)
        internal_points = np.array( list( reversed( points[1:] ) ), dtype=int )
        knots = self.x2[internal_points]

        self.added_knot_type = KNOT_NO_ADDITION
        self.soaring_peaks  = 0
        smoother = LSQUnivariateSpline(self.x2, self.log_y, knots)
        head_slice = slice(0,internal_points[0])
        head_x = self.x2[head_slice]
        head_y = self.log_y[head_slice] 
        head_len = len(head_y)

        # select 98-percentille point as peak-point to avoid abnormal max
        kth = int( head_len*98/100 )
        arg_hy98 = np.argpartition( head_y, kth )
        m = arg_hy98[kth]
        if self.debug_plot:
            print( 'head_len=', head_len, 'kth=', kth, 'm=', m )
        if m < MIN_NUM_HEAD_DROPS:
            gr_fwd = np.gradient( head_y[m:m+FWD_NUM_HEAD_DROPS] )
            gr_ratio = len( np.where( gr_fwd > 0 )[0] ) / FWD_NUM_HEAD_DROPS
            if self.debug_plot:
                print( 'gr_ratio=', gr_ratio, 'gr_fwd=', gr_fwd )
            # if it has a significant increasing part, try again for the next forward part
            if gr_ratio > 0.35:
                kth = int( ( head_len - m )*98/100 )
                arg_hy98 = np.argpartition( head_y[m:], kth )
                m = m + arg_hy98[kth]
                if self.debug_plot:
                    print( 'kth=', kth, 'm=', m )

        # if self.debug_plot:
        if False:
            from CanvasDialog   import CanvasDialog
            def func( fig ):
                ax = fig.add_subplot(111)
                ax.set_title( 'Peak Debug')
                ax.plot( head_x, head_y, 'o', markersize=3 )
                ax.plot( head_x[m], head_y[m], 'o', color='red', markersize=3  )
                ax.plot( head_x[0:m], head_y[0:m], 'o', color='orange', markersize=3  )
                ax.plot( head_x, smoother(head_x), color='green' )
                ax.set_xlim( ax.get_xlim() )
                ax.set_ylim( ax.get_ylim() )
                ax.plot( knots, smoother(knots), 'o', color='red', markersize=3)
                fig.tight_layout()
            dialog = CanvasDialog( "Debug" )
            dialog.show( func, figsize=(9,8), toolbar=True )

        self.stop = internal_points[0]  # note that this setting is not final

        self.fixed_guinier_start = get_setting( 'fixed_guinier_start' )
        if self.fixed_guinier_start == 1:
            m = get_setting( 'guinier_start_point' )
            self.guinier_start_point = m
            self.set_peak_info( m, smoother )

        else:

            self.set_peak_info( m, smoother )

            if m > 1:
                # pre_head_mean = np.average( head_y[0:m] )
                pre_head_mean   = np.percentile( head_y[0:m], 50 )
            else:
                pre_head_mean   = head_y[m]
            # print( 'pre_head_mean=', pre_head_mean, 'log_py=', self.log_py, 'head_y[0:m]=', head_y[0:m] )

            if m >= MIN_NUM_HEAD_DROPS and pre_head_mean < self.log_py:
                height_base = smoother(self.x2[0])
                height_p    = self.log_py  - height_base
                height_sp   = self.log_spy - height_base
                h_ratio = height_p/height_sp
                if height_sp > 0 and h_ratio >= TODO_DROPPING_RATIO:
                    self.peak_type  = '1-1'
                    # this indicates dropping head
                    knots = np.array( [self.px2] + list( knots ) )
                    self.added_knot_type = KNOT_DROPPING_HEAD
                    # print( 'knots=', knots )
                    smoother = LSQUnivariateSpline(self.x2, self.log_y, knots)
                    # why don't do set_peak_info?
                else:
                    self.peak_type  = '1-0'
            else:
                log_ym      = self.log_y[m]
                log_sym     = smoother( self.x2[m] )
                head_gap1   = log_ym - log_sym
                # head_gap2   = log_ym - smoother( self.x2[internal_points[0]] )
                if head_gap1 > 0:
                    self.peak_type  = '0-1'
                    m20 = min( len( self.x2 ) - 1, 20 )     # len( self.x2 )==15 for 20160426-9
                    head_gap2   = log_sym - smoother( self.x2[m20] )
                    # print( 'head_gap1=', head_gap1, 'head_gap2=', head_gap2 )
                    if head_gap2 > 0:
                        head_gap_ratio = head_gap1/head_gap2
                    else:
                        head_gap_ratio = TODO_SOARING_RATIO + 10
                        # i.e., always do find_better_peak

                    # print( 'head_gap_ratio=', head_gap_ratio )
                    self.head_gap_ratio = head_gap_ratio
                    if head_gap_ratio > TODO_SOARING_RATIO:
                        self.peak_type  = '0-1-1'
                        # case like pH6.136
                        m_try = self.find_better_peak( head_gap_ratio, m, head_x[m:], head_y[m:] )
                        if m_try is not None:
                            self.peak_type  = '0-1-1-1'
                            for t_ in range(2):
                                ret_smoother = self.try_better_peak( head_gap_ratio, m, m_try, knots, smoother, head_x, head_y )
                                if ret_smoother is not None:
                                    self.peak_type  = '0-1-1-1-1'
                                    smoother = ret_smoother
                                    self.added_knot_type = KNOT_SOARING_HEAD
                                    self.set_peak_info( m_try, smoother )
                                    break

                                if m_try < PEAK_FIND_RETRIABLE_LEN:
                                    break
                                else:
                                    # as in TSsome2-225
                                    m_try = self.find_better_peak( head_gap_ratio, m, head_x[m:m_try], head_y[m:m_try] )
                                    if m_try is None:
                                        break
                                if debug:
                                    print('retrying: for m_try=', m_try)
                else:
                    self.peak_type  = '0-0'

        self.log_sy = smoother(self.x2)
        self.log_sy_max = np.max( self.log_sy[0:internal_points[0]] )
        self.kx2 = smoother.get_knots()
        self.log_ky = smoother(self.kx2)
        self.sy = np.exp(self.log_sy)
        self.kx = np.sqrt(self.kx2)
        self.ky = np.exp(self.log_ky)
        self.stop = internal_points[0]

    def set_peak_info( self, m, smoother ):
        self.peak = m
        self.px  = self.x_[m]
        self.px2 = self.x2[m]
        self.log_py = self.log_y[m]
        self.log_spy = smoother( self.px2 )

    def find_better_peak(self, head_gap_ratio, start, head_x, head_y, debug=False):
        """
        note that we can't yet use self.basic_quality here
        because evaluate_basic_quality depends the peak info determined here
        """
        # scale x so that the slope be equal to -1 and rotate -pi/9
        # changed -pi/4 to -pi/6 for AhRR-124
        span_y = head_y[-1] - head_y[0]
        span_x = head_x[-1] - head_x[0]
        span_ratio = abs( span_y/span_x )
        hx = head_x * span_ratio

        # if self.debug_plot:
        if False:
            from CanvasDialog   import CanvasDialog
            def func( fig ):
                ax = fig.add_subplot(111)
                ax.set_title( 'hx, head_y')
                ax.plot( hx, head_y )
                fig.tight_layout()
            dialog = CanvasDialog( "Debug" )
            dialog.show( func, figsize=(9,8), toolbar=True )

        # consider changing angle to -np.pi/4 for Matsumura2-78
        # print( 'span_ratio=', span_ratio )
        angle = PEAK_FIND_ROTATION
        x, y = rotate( angle, hx, head_y )

        def get_partition_index( percentile ):
            kth = int( len(x)*percentile/100 )
            x_argp = np.argpartition( x, kth )
            return x_argp[0:kth+1]

        p_index = get_partition_index( PEAK_FIND_PERCENTILE )
        # print( 'p_index=', p_index )

        x_  = x[p_index]
        y_  = y[p_index]

        # center of gravity
        cx  = np.average( x_ )
        cy  = np.average( y_ )

        pos_type = None

        # for TRN1PY-195, head_gap_ratio==1.64, 'largest' should be avoided
        if head_gap_ratio > GOOD_SOARING_RATIO:
            x_width = np.max( x - cx )
            x_dev = np.std( x_ ) / x_width
            # print( 'x_dev=', x_dev, 'x_width=', x_width )
            if x_dev < UP_ADJUST_X_DEV_BNDRY:
                # in cases such as SUB_TRN1-160
                # largest y_
                y_  = y[p_index]
                k   = np.argmax( y_ )
                m   = p_index[k]
                pos_type = 'largest'

        if pos_type is None:
            # in cases such as pH7-243
            # find nearest to the center of gravity
            d2  = (x_ - cx)**2 + (y_ - cy)**2
            k   = np.argmin( d2 )
            m   = p_index[k]
            pos_type = 'nearest to center'

        # print( 'find_better_peak: m=', m )

        """
        span_rx = np.max( np.abs( x - x[m] ) )
        span_ry = np.max( np.abs( y - y[-1] ) )
        span_ratio = span_ry / span_rx
        print( 'span_ratio=', span_ratio )
        """

        if self.debug_plot:
            from molass_legacy.KekLib.CanvasDialog import CanvasDialog
            def func( fig ):
                ax = fig.add_subplot(111)
                ax.set_title( '%g° rotated x, y' % ( 180*angle/np.pi ) )
                ax.plot( x[0:m], y[0:m], 'o', markersize=3 )
                tx, ty = x[m], y[m]
                ax.plot( x[m:], y[m:], 'o', markersize=3 )
                for k in p_index:
                    ax.plot( x[k], y[k], 'o', color='orange', markersize=3 )
                ax.plot( cx, cy, 'o', color='yellow' )
                ax.plot( tx, ty, 'o', color='red' )
                xmin, xmax = ax.get_xlim()
                ymin, ymax = ax.get_ylim()
                xoffset = (xmax - xmin)*0.1
                yoffset = (ymax - ymin)*0.2
                if tx < xmin + xoffset:
                    xoffset *= -1
                    yoffset *= -0.5
                text = 'center of gravity for %d or lower percentile points' % PEAK_FIND_PERCENTILE
                ax.annotate( text, xy=( cx, cy ),
                    xytext=( cx - xoffset, cy + yoffset ), alpha=0.5,
                    arrowprops=dict( headwidth=5, width=1, facecolor='black', shrink=0, alpha=0.3),
                    # ha='center',
                    )
                text = pos_type + ' among %d or lower percentile points' % PEAK_FIND_PERCENTILE
                ax.annotate( text, xy=( tx, ty ),
                    xytext=( tx - xoffset, ty - yoffset ), alpha=0.5,
                    arrowprops=dict( headwidth=5, width=1, facecolor='black', shrink=0, alpha=0.3),
                    # ha='center',
                    )
                fig.tight_layout()
            dialog = CanvasDialog( "Debug" )
            dialog.show( func, figsize=(9,8), toolbar=True )

        try:
            slope, r_value = stats.linregress( x[m:], y[m:] )[LINREG_INDEX]
            # print( 'find_better_peak: (1) r_value=', r_value )
            min_r_value = None
            min_r_n     = None

            if r_value > PEAK_FIND_R_VALUE:
                # try modifying m to get an expected situation
                m_stop = max( start + 5, m - MAX_NUM_GOING_BACK )
                for n in range( m-1,  m_stop, -1):
                    slope, r_value = stats.linregress( x[n:], y[n:] )[LINREG_INDEX]
                    if self.debug_plot:
                        print( 'find_better_peak: (2)', [n], 'r_value=', r_value )
                    if r_value <= PEAK_FIND_R_VALUE:
                        if debug:
                            print( 'modified m=%d to %d' % ( m, n ) )
                        m = n
                        break
                    if min_r_value is None or r_value < min_r_value:
                        min_r_value = r_value
                        min_r_n     = n

            # print( 'find_better_peak: r_value=', r_value, 'min_r_value=', min_r_value )
            if ( r_value > PEAK_FIND_R_VALUE
                and min_r_value is not None and min_r_value > PEAK_FIND_R_VALUE_ALLOW ):
                # not appropriate because the linearity is not sufficient
                m = None
            else:
                bck_slope, bck_r_value, bck_stderr = self.compute_backward_r_value( r_value, m, head_x, head_y )
                rg = compute_rg( slope )
                bck_rg = compute_rg( bck_slope )
                rg_ratio = bck_rg / rg if rg > 0 else 1e10
                bck_error = bck_stderr/abs( head_y[0] - head_y[-1])
                # print( 'bck_r_value=', bck_r_value, 'rg_ratio=', rg_ratio, 'bck_error=', bck_error )
                self.rg_ratio_for_better_peak   = rg_ratio
                # use bck_stderr because we can't use self.basic_quality yet
                # checking bck_error should avoid misjudgement for 20160227-23, etc.
                if bck_slope == 0:  # for Matsumura2-78
                    pass
                elif bck_error > PEAK_FIND_QUALITY_BNDRY and rg_ratio > MIN_BETTER_RG_RATIO:
                    # this peak adoption is valid onlyfor
                    #       for bad-quality data
                    #   and with big rg_ratio
                    # e.g. in case AhRR-133
                    pass
                else:
                    if head_gap_ratio > GOOD_SOARING_RATIO:
                        if ( bck_r_value < CANDIDATE_R_VALUE        # neccesary e.g. for 20160227
                            or bck_r_value + BACK_R_VALUE_PENALTY < r_value ):
                            # do not adopt this peak because the backward section has better linearity
                            m = None
        except:
            if True:
                from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
                etb = ExceptionTracebacker()
                print( etb )
            m = None

        if m is not None:
            self.soaring_peaks  += 1
            if m > PEAK_FIND_TOO_LONG:
                # TODO: better decision
                # save this peak candidate for stop point in make_cadidate_pairs
                # i.e. it would be inappropriate if we exceed this point
                # examples for this case include 20160227-124
                self.too_long_peak_candidate = start + m
                if debug:
                    print( 'too_long_peak_candidate=',  self.too_long_peak_candidate)
                m = self.find_better_peak( head_gap_ratio, start, head_x[0:m], head_y[0:m] )
                if m is not None:
                    m += start
            else:
                pass

        # print( 'find_better_peak: finally m=', m )

        return m

    def try_better_peak(self, head_gap_ratio, m, m_try, knots, smoother, head_x, head_y, debug=False):
        """
        we can adopt this peak if it substantially changes the spline
        """
        px = self.x2[m_try]
        # print( 'try_better_peak: ', knots )

        if ( m_try - m < TRY_PEAK_MIN_OFFSET    # like Jun28-157
            or px > knots[0]                    # Interior knots t must satisfy Schoenberg-Whitney conditions
            ):
            return None

        knots = np.array( [px] + list( knots ) )
        # print( 'knots=', knots )
        new_smoother = LSQUnivariateSpline(self.x2, self.log_y, knots)

        hx  = head_x[m]
        # bm  = min( m+MIN_GUINIER_SIZE, m_try )
        bm  = m + MIN_GUINIER_SIZE
        bslice = slice(m, bm)
        bx_ = self.x2[bslice]
        by_ = self.log_y[bslice]
        slope_bck, _ = stats.linregress( bx_, by_ )[LINREG_INDEX]

        px  = self.x2[m_try]
        fx  = self.x2[m_try+MIN_GUINIER_SIZE]
        shy, spy, sfy = new_smoother( [hx, px, fx] )
        slope_fwd   = ( sfy - spy ) / ( fx - px )

        rg_bck      = compute_rg( slope_bck )
        rg_fwd      = compute_rg( slope_fwd )
        if rg_fwd > 0:
            rg_ratio    = rg_bck / rg_fwd
        else:
            rg_ratio    = PEAK_FIND_MIN_RG_RATIO * 10

        # compute both( bck and fwd ) monotonicity
        # for judging from only one of them is risky
        bck_monotonicity, bck_tangent = compute_monotonicity( head_y[m:m_try] )
        fwd_monotonicity, fwd_tangent = compute_monotonicity( head_y[m_try:] )
        monotonicity_ratio = bck_monotonicity / fwd_monotonicity if fwd_monotonicity > 0 else 0

        # compute the tangent of the angle difference
        tangent_diff = ( fwd_tangent - bck_tangent ) / ( 1 + fwd_tangent * bck_tangent )

        peak_fwd_qrg    = self.x[m_try+MIN_GUINIER_SIZE] * rg_fwd

        if self.debug_plot:
            print( 'm=', m, 'm_try=', m_try )
            print( 'rg_ratio=', rg_ratio )
            print( 'bck_monotonicity=', bck_monotonicity )
            print( 'fwd_monotonicity=', fwd_monotonicity )
            print( 'monotonicity_ratio=', monotonicity_ratio )
            print( 'bck_tangent=', bck_tangent )
            print( 'fwd_tangent=', fwd_tangent )
            print( 'tangent_diff=', tangent_diff )
            print( 'peak fwd qrg=', peak_fwd_qrg )

            from molass_legacy.KekLib.CanvasDialog   import CanvasDialog
            hy  = head_y[0]
            py  = self.log_y[m_try]

            def func( fig ):
                ax = fig.add_subplot(111)
                ax.set_title( 'try_better_peak')
                ax.plot( head_x, head_y, 'o', markersize=3 )
                ax.plot( head_x, smoother(head_x) )
                ax.plot( head_x, new_smoother(head_x) )
                ax.plot( hx, hy, 'o', color='red' )
                ax.plot( hx, smoother(hx), 'o', color='yellow' )
                ax.plot( hx, new_smoother(hx), 'o', color='pink' )
                ax.plot( px, py, 'o', color='orange' )
                ax.plot( px, new_smoother(px), 'o', color='orange' )
                fig.tight_layout()
            dialog = CanvasDialog( "Debug" )
            dialog.show( func, figsize=(9,8), toolbar=True )

        if peak_fwd_qrg > PEAK_FIND_QRG_LIMIT:
            # this case, as in Open01-232, is useless because no cadidates would be found
            self.try_ret = -1
            return None

        # the order of the following tests are relevant
        # so that you need to be very careful in maintenancer

        if fwd_monotonicity > MONOTONICITY_FWD_OK:
            if tangent_diff > TANGENT_DIFF_BOUNDARY:
                self.try_ret = 1
                return new_smoother

            cv = curvature_curve( head_x, head_y, spline=new_smoother )
            cv_try = cv[m_try]
            # print( 'cv_try=', cv_try )
            if cv_try > PEAK_FIND_MIN_CURVATURE:
                # as in SUB_TRNPY-193
                self.try_ret = 10
                return new_smoother
            else:
                self.try_ret = -10
                return None

        if fwd_monotonicity < QUALITY_BOUNDARY_LOW:
            if rg_ratio < PEAK_FIND_RG_RATIO_LOW:
                # like Matsumura-166, 12/2-81
                self.try_ret = -11
                return None

            if rg_ratio > PEAK_FIND_MIN_RG_RATIO:
                # like Nov24-209
                self.try_ret = 11
                return new_smoother

            # compute possible_size(=num points) at this point from the spline
            psy = new_smoother( px )
            fx  = self.x2[m_try + MIN_GUINIER_SIZE]
            fsy = new_smoother( fx )
            slope = ( fsy - psy )/( fx - px )
            rg = compute_rg( slope )
            max_q = 1.3/rg if rg > 0 else 1
            possible_size = max( 0, max_q - self.x[m_try] ) / self.dx
            if debug:
                print( 'possible_size=', possible_size, 'rg=', rg, 'max_q=', max_q, 'self.x[m_try]=', self.x[m_try], 'self.dx=', self.dx )
            if possible_size < MIN_GUINIER_SIZE:
                # like ?
                self.try_ret = -12
                return None
            else:
                # like AhRR-10, AhRR-134,
                # before checking ratio, too bad fwd_monotonicity should be avoided
                # note that this includes the case like HIF-74 where fwd_monotonicity==0
                self.try_ret = 12
                return new_smoother

        if monotonicity_ratio > GOOD_MONOTONICITY_RATIO:
            # there is a good possibility in backward part
            # so, adopting this peak is not a good idea
            self.try_ret = -2
            return None

        if ( fwd_monotonicity < QUALITY_BOUNDARY_QRG
            and rg_ratio < PEAK_FIND_RG_RATIO_LOW ):
            # for Matsumura-258
            self.try_ret = -3
            return None

        if fwd_monotonicity > QUALITY_BOUNDARY_QRG:
            self.try_ret = 2
            return new_smoother

        if fwd_monotonicity < QUALITY_BOUNDARY_LOW and rg_ratio < PEAK_FIND_RG_RATIO_LOW:
            # for 12/2-113
            self.try_ret = -4
            return None

        if head_gap_ratio <=  GOOD_SOARING_RATIO:
            # for test_SUB_TRN1PY-195
            self.try_ret = 3
            return new_smoother

        if  bck_monotonicity > QUALITY_BOUNDARY_BEST:
            # even if monotonicity_ratio is not enough
            # very good bck_monotonicity suggest a good possibility
            # assert False        # to see if this case be included monotonicity_ratio > GOOD_MONOTONICITY_RATIO
            # return None
            # for Ald-136, returning None is irrelevant
            pass

        if rg_ratio > PEAK_FIND_MIN_RG_RATIO:
            # too big rg ratio seems to indicate unacceptable soaring
            self.try_ret = 4
            return new_smoother

        self.try_ret = -5
        return None

    def compute_backward_r_value( self, r_value, m, head_x, head_y ):
        size = min( 40, len(head_x) - m )

        start = max( 0, m-size )
        slice_ = slice( start, m )

        if m - start < MIN_GUINIER_SIZE:
            return 0, 0, 1e5

        if False:
            from CanvasDialog   import CanvasDialog
            def func( fig ):
                ax = fig.add_subplot(111)
                ax.set_title( 'compute_backward_r_value')
                ax.plot( head_x, head_y, 'o' )
                ax.plot( head_x[slice_], head_y[slice_] )
                fig.tight_layout()
            dialog = CanvasDialog( "Debug" )
            dialog.show( func, figsize=(9,8), toolbar=True )

        reg_ret = stats.linregress( head_x[slice_], head_y[slice_] )
        slope, r_value = reg_ret[LINREG_INDEX]
        stderr = reg_ret[LINREG_STDERR]
        return slope, r_value, stderr

    def evaluate_basic_quality(self, debug=False):
        log_y = self.log_y[self.peak:self.stop]

        if len(log_y) < MIN_GUINIER_SIZE:
            self.basic_quality = None
            if debug:
                self.logger.warning("basic_quality is set to None due to %d < %d", len(log_y), MIN_GUINIER_SIZE)
            return

        height = log_y[0] - log_y[-1]
        if height > 0:
            ag_sum = np.sum( np.abs( np.gradient( log_y ) ) )
            self.ag_ratio = ag_sum / height
            self.basic_quality = np.exp( MONOTONICITY_SCALE *( 1 - self.ag_ratio ) )
        else:
            self.basic_quality = 0

        self.worst_quality = self.basic_quality < QUALITY_BOUNDARY_WORST

    def make_cadidate_pairs( self, qrg_limit ):
        # print( 'make_cadidate_pairs: qrg_limit=', qrg_limit )

        if self.worst_quality:
            wide_allow  = 0.5
        elif self.basic_quality > QUALITY_BOUNDARY_QRG:     # 0.5
            wide_allow  = 0.3
        elif self.basic_quality < QUALITY_BOUNDARY_LOW:     # 0.02
            wide_allow  = 1.0
        else:
            wide_allow  = 0.5

        qrg_limit_wide = qrg_limit + wide_allow

        if self.anim_data:
            self.anim_ag_lines = []

        if self.worst_quality:
            log_y_s = self.log_sy
        else:
            log_y_s = self.log_y

        if self.basic_quality < QUALITY_BOUNDARY_LOW:
            log_y_e = self.log_sy
        else:
            log_y_e = self.log_y

        if self.too_long_peak_candidate is None:
            loop_stop = self.stop
        else:
            loop_stop = self.too_long_peak_candidate

        # print( 'make_cadidate_pairs: loop_stop=', loop_stop )

        pairs = []
        if self.fixed_guinier_start == 1:
            start_limit = self.peak + 1
        else:
            start_limit = loop_stop - MIN_GUINIER_SIZE
            if self.x2[start_limit] > MAX_START_Q2:
                # identify these cases
                max_start   = bisect_right( self.x2, MAX_START_Q2 )
                start_limit = max( self.peak + MIN_START_FROM_PEAK, max_start )

        # print( 'peak=', self.peak, 'start_limit=', start_limit )

        for start in range( self.peak, start_limit ):
            end     = start + MIN_GUINIER_SIZE
            e_slice = slice(end, self.stop)
            # print( 'start, end, stop=', start, end, self.stop )
            slope   = ( log_y_s[start] - log_y_e[e_slice] ) / ( self.x2[start] - self.x2[e_slice] )
            neg_i   = slope < 0
            neg_i_w = np.where( neg_i )[0]
            if len(neg_i_w) == 0:
                continue

            qrg     = self.x[e_slice][neg_i] * np.sqrt( -3*slope[neg_i] )
            try:
                j   = np.where( qrg < qrg_limit_wide )[0][-1]
            except:
                if False:
                    from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
                    etb = ExceptionTracebacker()
                    print( etb )
                continue

            stop = end + neg_i_w[j] + 1
            pairs.append( [ start, stop ] )

            if self.anim_data:
                to = stop - 1
                self.anim_ag_lines.append(  [ (self.x2[start], self.x2[to] ), (self.log_y[start], self.log_y[to]) ] )

            if self.qrg_stop is None or stop > self.qrg_stop:
                self.qrg_stop       = stop
                self.guinier_stop   = stop  # this usually will be improved later

        if self.anim_data:
            self.anim_guinier_stop  = self.qrg_stop

        # print( 'pairs=', pairs )
        return pairs

    def evaluate_interval( self, i, start, stop_a, qrg_limit, qrg_allow ):

        qrg_limit_ = qrg_limit + qrg_allow

        candidate_count = 0
        for stop in range( stop_a, start + MIN_GUINIER_SIZE, -1 ):
            size = stop - start
            if size < MIN_GUINIER_SIZE:
                break

            slice_ = slice( start, stop )

            x_ = self.x2[slice_]
            y_ = self.log_y[slice_]
            slope, intercept, r_value, p_value, stderr = reg_result = stats.linregress( x_, y_ )

            candidate_rec = None
            if slope <= 0:
                rg = compute_rg( slope )
                x = self.x[stop-1]
                qrg = x*rg
                # print( '(start, stop)=', (start, stop), 'qrg=', qrg, 'r_value=', r_value, 'end_consistency=', end_consistency )
                if qrg < qrg_limit_ or self.anim_data:
                    end_consistency = compute_end_consistency( start, stop, self.x2, self.log_y )
                    fwd_consistency = compute_fwd_consistency( start, stop, self.x2, self.log_y )
                    if self.worst_quality:
                        spline_slope = (self.log_sy[start] - self.log_sy[stop-1])/(x_[0] - x_[-1])
                        spline_rg = compute_rg( spline_slope )
                    else:
                        spline_rg = None
                    if USE_GUINIER_PLOT_LENGTH:
                        length = np.sqrt( (x_[0] - x_[-1])**2 + (y_[0] - y_[-1])**2 )
                    else:
                        length = size
                    interval_score, score_vector, score_array = evaluate_guinier_interval( self.basic_quality, self.px, self.dx, rg, length, r_value, end_consistency, fwd_consistency, spline_rg, return_vector=True )
                    candidate_rec = [ i, start, stop, reg_result, rg, end_consistency, interval_score, score_vector ]
                    if self.anim_data:
                        # anim_score = interval_score if qrg < qrg_limit_ else 0
                        anim_score = interval_score
                        anim_rec = [ start, stop, x_, y_, slope, intercept, anim_score, score_vector, rg ]
                        self.anim_cand_list.append( anim_rec )
                else:
                    interval_score = 0
            else:
                interval_score = 0
                qrg = 999

            if qrg < qrg_limit_ and candidate_rec is not None:
                if self.score is None or interval_score > self.score:
                    self.score     = interval_score
                    self.candidate = candidate_rec
                    if self.anim_data:
                        # evaluate_guinier_interval( self.basic_quality, self.px, self.dx, rg, length, r_value, end_consistency, fwd_consistency, spline_rg, debug=True )
                        self.best_history_dict[( start, stop )] = [ score_vector, rg ]
                candidate_count += 1

                if x_[0] < SMALL_Q2_BOUNDARY:
                    if self.smallq_score is None or interval_score > self.smallq_score:
                        self.smallq_score       = interval_score
                        self.smallq_candidate   = candidate_rec

                if candidate_count > 3:
                    break

    def guinier_interval( self, qrg_limit=1.3 ):
        """
        q*Rg < 1.3
        log_y = log(G) - x**2 * Rg**2/3
        slope = -Rg**2/3
        Rg = sqrt(-3*slope)
        """
        if self.anim_data:
            self.anim_cand_list = []

        # TODO: determine qrg_allow thoeoretically
        if self.worst_quality:
            qrg_allow = 1.0
        elif self.basic_quality > QUALITY_BOUNDARY_QRG:
            qrg_allow = 0
        elif self.basic_quality < QUALITY_BOUNDARY_LOW:
            qrg_allow = 0.5
        else:
            qrg_allow = 0.3     # required for 20160227-4

        self.guinier_stop   = self.stop  # this usually will be improved later
        self.candidate = None
        self.score     = None
        self.smallq_candidate   = None
        self.smallq_score       = None

        lq_candidate_count = 0
        wst_candidate_count = 0
        pairs = self.make_cadidate_pairs( qrg_limit )
        for i, pair in enumerate(pairs):
            self.evaluate_interval( i, pair[0], pair[1], qrg_limit, qrg_allow )
            if self.worst_quality:
                if self.candidate is not None:
                    wst_candidate_count += 1
                    if wst_candidate_count >= WST_AT_LEAST_TRIALS:
                        break
            elif self.basic_quality < QUALITY_BOUNDARY_LOW:
                if self.candidate is not None:
                    lq_candidate_count += 1
                    if lq_candidate_count >= LQ_AT_LEAST_TRIALS:
                        break

        if self.candidate is not None:
            # print( 'self.candidate=', self.candidate[0:3], self.candidate[4] )
            pass

        if self.candidate is None:
            self.set_guinier_result( self.make_low_quality_rec() )
        else:
            candidate = self.candidate
            start = candidate[1]
            if self.x2[start] >= SMALL_Q2_BOUNDARY:
                if self.smallq_score is not None:
                    score_ratio = self.smallq_score / self.score
                    if score_ratio > SCORE_RATIO_BOUNDARY:
                        candidate   = self.smallq_candidate

            self.set_guinier_result( candidate )

    def make_low_quality_rec(self, debug=False):
        if debug:
            print( 'making low_quality_rec' )
        start = 0
        stop = self.guinier_stop

        if self.qrg_stop is None:
            self.qrg_stop = stop
        if self.anim_guinier_stop is None:
            self.anim_guinier_stop = self.qrg_stop

        slice_ = slice( start, stop )
        x_ = self.x2[slice_]
        y_ = self.log_y[slice_]
        reg_result = stats.linregress( x_, y_ )
        return [ None, start, stop, reg_result, 0, 0, 0, ZERO_SCORE_VECTOR ]

    def set_guinier_result( self, candidate_rec ):
        self.guinier_start  = candidate_rec[1]
        self.guinier_stop   = candidate_rec[2]
        self.guinier_length = self.guinier_stop - self.guinier_start
        qunier_index = [ self.guinier_start, self.guinier_stop-1 ]
        self.guinier_x = self.x2[qunier_index]
        slope, intercept = candidate_rec[3][0:2]
        self.guinier_y = slope * self.guinier_x + intercept
        self.Rg = Rg = candidate_rec[4]
        self.Iz = np.exp( intercept )
        fr_ = self.guinier_start
        to_ = self.guinier_stop - 1
        self.min_q = self.x[fr_]
        self.min_qRg = self.min_q*Rg
        self.max_q = self.x[to_]
        self.max_qRg = self.max_q*Rg
        end_consistency_ = candidate_rec[5]

        e_size = END_GUINIER_SIZE*2
        if self.basic_quality < QUALITY_BOUNDARY_WORST or self.guinier_length < e_size:
            # as in 20160227-66
            # print( 'guinier_length=',  self.guinier_length )
            size_eval_limit         = ( min( e_size, self.guinier_length ) / e_size )**2
            quality_eval_limit      = min( self.basic_quality, QUALITY_BOUNDARY_WORST) / QUALITY_BOUNDARY_WORST
            eval_limit  = size_eval_limit * quality_eval_limit
            # print( 'eval_limit=', eval_limit )
            self.end_consistency    = min( eval_limit, end_consistency_ )
        else:
            self.end_consistency = end_consistency_
        self.r_value    = candidate_rec[3][2]
        self.score      = candidate_rec[6]
        self.score_vector = candidate_rec[7]

        if False:
            ret = compute_end_consistency( self.guinier_start, self.guinier_stop, self.x2, self.log_y, debug=True )
            print( 'ret=', ret )

        # print( 'set_guinier_result: r_value=', self.r_value, 'score=', self.score, 'max_qRg=', self.max_qRg, 'end_consistency=', self.end_consistency  )

    def set_guinier_null_result( self ):
        if False:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            raise RuntimeError( etb )
        self.guinier_x  = None
        self.guinier_y  = None
        self.Rg         = None
        self.Iz         = None
        self.min_q      = None
        self.min_qRg    = None
        self.max_q      = None
        self.max_qRg    = None
        self.end_consistency = 0
        self.basic_quality = 0
        self.score = 0
        self.score_vector = ZERO_SCORE_VECTOR



# -*- coding: utf-8 -*-
"""

    ファイル名：   settings.py

    処理内容：

        設定情報

"""
from __future__ import division, print_function, unicode_literals

from PersistentInfo     import PersistentInfo
from SAnglerMask        import SAnglerMask

# 永続的な記憶をオブジェクト化する（メモリに取り込む）。
setting_file    = 'settings.dump'
setting_info    = PersistentInfo( setting_file )
settings_       = setting_info.get_dictionary()
mask_           = None

ITEM_DEFAULTS = {
    'in_folder'         : None,
    'log_file'          : None,
    'mask_file'         : None,
    'adj_folder'        : None,
    'syn_folder'        : None,
    'op_option'         : 'MANUAL',
    'watch_interval'    : 180,
    }

def reload_settings():
    global setting_info
    global settings_

    setting_info    = PersistentInfo( setting_file )
    settings_       = setting_info.get_dictionary()

def clear_settings():
    global settings_
    settings_ = {}
    setting_info.set_dictionary( settings_ )

def get_setting( item ):
    assert( item in ITEM_DEFAULTS )
    value = settings_.get( item )
    if value == None:
        value = ITEM_DEFAULTS.get( item )
        set_setting( item, value )
    return value

def set_setting( item, value ):
    assert( item in ITEM_DEFAULTS )
    settings_[ item ] = value

def temporary_settings_begin():
    global settings_save
    settings_save = settings_.copy()

def temporary_settings_end():
    global settings_save
    global settings_
    settings_ = settings_save
    setting_info.set_dictionary( settings_ )

def save_settings():
    setting_info.save()

def set_mask( filepath ):
    global mask_

    try:
        mask_ = SAnglerMask( filepath )
        return True
    except Exception as err:
        return False

def get_mask():
    return mask_



"""
    Optimizer.GuinierDeviation.py

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
import logging
from bisect import bisect_right
import numpy as np
from scipy.stats import linregress
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from importlib import reload
import molass_legacy.GuinierTools.RgCurveUtils
reload(molass_legacy.GuinierTools.RgCurveUtils)
from molass_legacy.GuinierTools.RgCurveUtils import get_connected_curve_info, convert_to_milder_qualities, VALID_BASE_QUALITY

USE_NORMALIZED_RMSD_FOR_RGCURVES = True

VERY_SMALL_VALUE = 1e-10
GN_DEV_LOWER_LIMIT = -2     # exceeding this to lower values may result in abnormal states
RG_DEV_LOWER_LIMIT = -2     # same as above
MAXIMIZE_ADJUST = -1.5      # this value balances the optimization among [Guinier_deviation, SEC_conformance, FIT_max]
                            # i.e., the greater this value, the more weight for Guinier_deviation 
NEGATIVE_SLOPE_SCALE = 1e3
QRG_UPPER_BOUND = 1.3
USE_GUINIER_RGS = True
BAD_RG_VALUE = -1
MIN_RG = 5                  # to avoid abnormal situations as in 20191118_4 with 4 components by STC_MP where Rg < 5Å

class GuinierDeviation:
    def __init__(self, qv, rg_curve, rg_params, region_limit=0, composite=None):
        self.logger = logging.getLogger(__name__)
        self.qv = qv
        self.rg_curve = rg_curve
        x_, y_, rgv, qualities, valid_bools = get_connected_curve_info(rg_curve)
        self.x_ = x_
        self.y_ = y_
        valid_bool_seg = valid_bools[1]
        self.rgv = rgv[valid_bool_seg]  # valid_bools[1] : valid_bool_seg
        self.mqualities = convert_to_milder_qualities(qualities)
        self.max_mask = self.mqualities > VALID_BASE_QUALITY
        self.weights = self.mqualities[valid_bool_seg]
        self.valid_bools = valid_bools
        self.valid_size = len(np.where(valid_bool_seg)[0])

        assert composite is not None
        self.composite = composite
        self.update_region(rg_params, region_limit) # this only roughly determines the region,
                                                    # which is to be refined later by update_region with P assigned
                                                    # in BasicOptimizer.update_guinier_region

    def update_region(self, rg_params, region_limit, P=None, M=None, E=None, debug=False):
        i0 = bisect_right(self.qv, region_limit)
        i1_rough_limit = len(self.qv)//2    # to avoid abnormal situations as in 20191118_4 with 4 components by STC_MP where Rg < 5Å
        self.gslices = []
        self.qv2s = []
        self.low_vecs = []
        self.rgs = np.ones(len(rg_params)) * np.nan
        self.izs = np.ones(len(rg_params)) * np.nan
        if P is None:
            pass
        else:
            # see also Optimizer.LrfExporter.XrLrfResult
            X = np.linalg.pinv(M) @ P       # M or M_
            Ep = np.sqrt((E**2) @ (X**2))
        valid_rgs = self.composite.get_valid_rgs(rg_params)
        if debug:
            data_list = []
            qualities = []
        for k, rg in enumerate(valid_rgs):
            if P is None:
                q = QRG_UPPER_BOUND/max(MIN_RG, rg)
                i1 = min(i1_rough_limit, max(i0 + 2, bisect_right(self.qv, q)))
                gslice = slice(i0, i1)
                iz = 1  # dummy, immediately replaced by the following values
            else:
                y = P[:,k]
                e = Ep[:,k]
                data = np.array([self.qv, y, e]).T
                sg = SimpleGuinier(data)
                if sg.Rg is None:
                    q = QRG_UPPER_BOUND/max(MIN_RG, rg)
                    i1 = min(i1_rough_limit, max(i0 + 2, bisect_right(self.qv, q)))
                    gslice = slice(i0, i1)
                    iz = y[i0]
                else:
                    self.rgs[k] = sg.Rg
                    i0_ = bisect_right(self.qv, sg.min_q)
                    i1 = bisect_right(self.qv, sg.max_q)
                    gslice = slice(i0_, i1)
                    iz = sg.Iz
                if debug:
                    data_list.append(data)
                    qualities.append(sg.basic_quality)
            self.gslices.append(gslice)
            self.qv2s.append(self.qv[gslice]**2)
            self.izs[k] = iz
            self.low_vecs.append(np.ones(gslice.stop - gslice.start) * VERY_SMALL_VALUE)

        if debug:
            print("rg_params=", rg_params)
            print("self.gslices=", self.gslices)
            qv2w = self.qv**2

            max_stop = np.max([slice_.stop for slice_ in self.gslices])
            """
            q*Rg = 1.3
            q = 1.3/Rg
            """
            min_rg = np.min(valid_rgs)
            lim_q = 1.8/min_rg
            max_stop = max(max_stop, bisect_right(self.qv, lim_q))

            indeces = [str(i) for i in range(len(qualities))]
            with plt.Dp():
                fig, (ax0, ax1, ax2) = plt.subplots(ncols=3, figsize=(18,5))
                fig.suptitle("Debug Plot for update_region", fontsize=20)
                ax0.set_title("Linear Plot", fontsize=16)
                ax1.set_title("Guinier Plot", fontsize=16)
                ax2.set_title("Rg Basic Qualities", fontsize=16)
                k = 0
                for gslice, qv2, data in zip(self.gslices, self.qv2s, data_list):
                    lin_y = data[:,1]
                    ax0.plot(self.qv, lin_y, label="$R_g$=%.3g" % self.rgs[k])
                    log_y = np.log(lin_y)
                    ax1.plot(qv2w[0:max_stop], log_y[0:max_stop], label="$R_g$=%.3g" % self.rgs[k])
                    ax1.plot(qv2, log_y[gslice], "o", color="yellow")
                    k += 1
                ax0.legend()
                ax1.legend()
                ax2.bar(indeces, qualities)
                fig.tight_layout()
                ret = plt.show()
                if not ret:
                    return

        regions = [self.qv[[gslice.start, gslice.stop]] for gslice in self.gslices]
        self.logger.info("updated regions=%s", str(regions))
        if P is not None:
            self.logger.info("updated rgs=%s", str(self.rgs))

        return True

    def compute_deviation(self, P, Cxr, rg_params, valid_components=None, return_details=False, debug=False):
        if valid_components is None:
            valid_components = np.ones(len(rg_params), dtype=bool)
        dev1 = 0
        dev2 = 0
        nc = self.composite.get_num_substantial_components()    # excluding the baseline
        valid_rgs = self.composite.get_valid_rgs(rg_params)
        n = 0
        lrf_rgs = []
        for k, p in enumerate(P.T[0:nc]):
            if not valid_components[k]:
                # continue
                pass

            x = self.qv2s[k]
            y = np.log(np.max([self.low_vecs[k], p[self.gslices[k]]], axis=0))
            slope, intercept, r_value, p_value, std_err = linregress(x, y)

            if slope < 0:
                rg = np.sqrt(-3*slope)
            else:
                rg = BAD_RG_VALUE                   # to avoid nan values
                std_err *= NEGATIVE_SLOPE_SCALE     # indicating it is undesirable

            lrf_rgs.append(rg)

            # print([k], std_err, rg, valid_rgs[k])
            dev1 += std_err**2
            if USE_GUINIER_RGS:
                if not np.isnan(self.rgs[k]):
                    dev2 += (self.rgs[k] - valid_rgs[k])**2
            else:
                dev2 += (rg - valid_rgs[k])**2
            n += 1

        gdev = max(GN_DEV_LOWER_LIMIT, np.log10(dev1)) + MAXIMIZE_ADJUST
        if n == 0:
            # to avoid division by zero in dev2/n
            n = 1
        rdev = max(RG_DEV_LOWER_LIMIT, np.log10(dev2/n)) + MAXIMIZE_ADJUST
        ret_val = ((gdev + rdev)/2 + max(gdev, rdev))/2
        if debug:
            from importlib import reload
            def rg_deviation_inspect():
                import molass_legacy.GuinierTools.RgCurveUtils
                reload(molass_legacy.GuinierTools.RgCurveUtils)
                from .RgCurveUtils import rg_deviation_inspect_impl
                rg_deviation_inspect_impl(self, valid_components, rg_params, lrf_rgs)

            def compare_dist_measures():
                import molass_legacy.GuinierTools.CompareGdevMeasures
                reload(molass_legacy.GuinierTools.CompareGdevMeasures)
                from .CompareGdevMeasures import compare_gdev_measures_impl
                print("ret_val=", ret_val)
                compare_gdev_measures_impl(self, Cxr, rg_params, valid_components)

            extra_button_specs=[
                ("inspect", rg_deviation_inspect),
                ("compare", compare_dist_measures),
                ]

            with plt.Dp(button_spec=["OK", "Cancel"], extra_button_specs=extra_button_specs):
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("compute_deviation")
                ax1.bar(["dev1", "dev2"], [dev1, dev2])
                ax2.bar(["adjust", "glimit", "gdev", "rlimit", "rdev", "ret_val"], [MAXIMIZE_ADJUST, GN_DEV_LOWER_LIMIT, gdev, RG_DEV_LOWER_LIMIT, rdev, ret_val])
                fig.tight_layout()
                ret = plt.show()
            if not ret:
                return

        if return_details:
            return gdev, rdev
        else:
            return ret_val

    def compute_rgcurve_deviation(self, Cxr, rg_params, adjust=0, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.GuinierTools.RgCurveUtils
            reload(molass_legacy.GuinierTools.RgCurveUtils)
        from .RgCurveUtils import get_reconstructed_curve
        
        rrgv = get_reconstructed_curve(self.valid_size, self.valid_bools, Cxr, rg_params)
        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("get_reconstructed_curve: debug")
                ax.plot(self.rgv, label='guinier measured')
                ax.plot(rrgv, label='reconstructed')
                ax.legend()
                fig.tight_layout()
                plt.show()

        if USE_NORMALIZED_RMSD_FOR_RGCURVES:
            if debug:
                import molass_legacy.Distance.NormalizedRmsd
                reload(molass_legacy.Distance.NormalizedRmsd)
            from molass_legacy.Distance.NormalizedRmsd import normalized_rmsd
            return normalized_rmsd(self.rgv, rrgv, weights=self.weights, adjust=adjust)
        else:
            from molass_legacy.Distance.FrobeniusXdiffmax import frobenius_xdiffmax
            return frobenius_xdiffmax(self.rgv, rrgv, adjust=adjust, max_mask=self.max_mask)


"""
    Optimizer.FullOptInput.py

    Copyright (c) 2021-2023, SAXS Team, KEK-PF

    First note that FullOptInput is constructed in the both processes, namely
        ・ monitor process and
        ・ optimizer process,
    as detailed below.

    (1) in the monitor process

        ・ OptimizerUtils.show_peak_editor_impl
            ・ PeakEditor.prepare_rg_curve → FullOptInput.get_dsets()
            ・ InitialInfo.__init__,
                ・ dsets = pe.get_dsets()
                ・ dsets.relocate_rg_folder()
        ・ show_result_folder_selector_impl → FullOptInput.get_dsets()

    (2) in the optimizer process

        ・ OptimizerMain → FullOptInput.get_dsets()
"""
import os
import logging
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

class FullOptInput:
    def __init__(self, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.update_attributes(**kwargs)
        if self.trimming_txt is not None:
            self.restore_trimming_info()

    def update_attributes(self, **kwargs):
        in_folder = kwargs.pop('in_folder', None)
        if in_folder is None:
            in_folder = get_setting('in_folder')
        self.in_folder = in_folder
        self.dsets = kwargs.pop('dsets', None)
        self.rg_folder = kwargs.pop('rg_folder', None)
        self.trimming_txt = kwargs.pop('trimming_txt', None)
        self.sd = kwargs.pop('sd', None)
        self.corrected_sd = kwargs.pop('corrected_sd', None)
        self.mapper = kwargs.pop('mapper', None)
        self.mapped_params = kwargs.pop('mapped_params', None)
        self.peak_params = kwargs.pop('peak_params', None)

    def get_sd(self, in_folder=None):
        if in_folder is None:
            in_folder = self.in_folder

        sd = self.sd
        if sd is None :
            if in_folder is None:
                # i.e., no input data info
                sd = None
            else:
                sd, corrected_sd = self.get_sd_from_folder(in_folder)
                self.sd = sd
                self.corrected_sd = corrected_sd
        return sd

    def get_dsets(self, in_folder=None, progress_cb=None, compute_rg=False, possibly_relocated=False, debug=False):
        sd = self.get_sd(in_folder=in_folder)
        if sd is None:
            return None

        dsets = self.dsets
        if dsets is None:
            if debug:
                from importlib import reload
                import Optimizer.OptDataSets
                reload(Optimizer.OptDataSets)
            from .OptDataSets import OptDataSets
            corrected_sd = self.corrected_sd
            self.dsets = dsets = OptDataSets(sd, corrected_sd,
                                    rg_folder=self.rg_folder,
                                    progress_cb=progress_cb,
                                    compute_rg=compute_rg,
                                    possibly_relocated=possibly_relocated,
                                    )

        return dsets

    def get_sd_from_folder(self, in_folder):
        from .TheUtils import get_sd_from_folder_impl
        sd, corrected_sd, self.treat = get_sd_from_folder_impl(in_folder, self.logger)
        return sd, corrected_sd

    def get_base_curve(self):
        return self.treat.get_base_curve()

    def get_mapped_params(self):
        if self.mapped_params is not None:
            return self.mapped_params

        if self.mapper is None:
            self.sp.get_corrected_sd(proxy=False)
            self.mapper = self.sp.mapper
        mapper = self.mapper
        mapped_info = mapper.get_mapped_info()
        opt_results = mapped_info.opt_results
        A, B, _ = np.average(np.array(opt_results), axis=0)
        num_peaks_mapping = len(opt_results)
        self.logger.info("num_peaks_mapping=%d", num_peaks_mapping)
        return A, B, num_peaks_mapping

    def restore_trimming_info(self):
        from molass_legacy.Trimming import restore_trimming_info_impl
        restore_trimming_info_impl(self.trimming_txt, self.logger)



"""
    Models.Stochastic.RoughGuess.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from scipy.spatial import cKDTree
from matplotlib.widgets import Slider, CheckButtons
import molass_legacy.KekLib.DebugPlot as plt
from Experiment.ColumnTypes import get_all_poresizes
from SecTheory.BasicModels import robust_single_pore_pdf
from molass_legacy.Models.ModelUtils import compute_raw_moments1, compute_area_props
from molass_legacy.Models.Stochastic.ParamLimits import MAX_PORESIZE
from importlib import reload
import Models.Stochastic.MonoporeUtils
reload(Models.Stochastic.MonoporeUtils)
from molass_legacy.Models.Stochastic.MonoporeUtils import compute_monopore_curves, plot_monopore_curves, adjust_scales_to_egh_params_array

NUM_STAYS_LIST = [1000, 2000, 3000, 4000, 5000]
M1_WEIGHT = 9
M2_WEIGHT = 0.5
M3_WEIGHT = 0.5

def guess_monopore_params_roughtly(x, y, model, peaks, peak_rgs, props, egh_moments_list, debug=False):
    if debug:
        print("props=", props)
        for k, moments in enumerate(egh_moments_list):
            print([k], moments)

    if False:
        from molass_legacy.Models.ModelUtils import data_dump
        data_dump(x, y, peaks, prefix="roughtly-")

    me = 1.5
    mp = 1.5
    max_rg = np.max(peak_rgs)
    poresizes = np.asarray(get_all_poresizes())
    possible_poresizes = poresizes[poresizes > max_rg + 20]
    dev_list = []
    for poresize in possible_poresizes:
        for N in NUM_STAYS_LIST:
            def rough_objective(p, debug=False):
                T_, x0_ = p
                dev_list = []
                cy_list = []
                for k, (rg, M) in enumerate(zip(peak_rgs, egh_moments_list)):
                    rho = min(1, rg/poresize)
                    M1_ = x0_ + N * T_ * (1 - rho)**(me+mp)
                    M2_ = 2 * N * T_**2 * (1 - rho)**(me+2*mp)
                    dev = M1_WEIGHT*(M1_ - M[0])**2 + M2_WEIGHT*(M2_ - M[1])**2
                    dev_list.append(dev)
                    ni_ = N * (1 - rho)**me
                    ti_ = T_ * (1 - rho)**mp
                    cy = props[k] * robust_single_pore_pdf(x - x0_, ni_, ti_)
                    cy_list.append(cy)
                mdev = np.sum(np.asarray(dev_list)*props)
                if debug:
                    print("poresize, N, T, x0=", poresize, N, T_, x0_)
                return mdev
            init_T = 0.5
            init_x0 = 0
            bounds = ((0, 10), (-500, 200))
            res = minimize(rough_objective, [init_T, init_x0], bounds=bounds, method="Nelder-Mead")
            dev_list.append((poresize, N, *res.x, res.fun))
    dev_list = np.asarray(dev_list)

    # remove outliers so that we can get a better view in 3D debug plot
    fv_temp = dev_list[:,-1]
    k8 = int(0.8*len(fv_temp))
    pp = np.argpartition(fv_temp, k8)
    p8 = sorted(pp[0:k8])
    fv_x = dev_list[p8,0]
    fv_y = dev_list[p8,1]
    fv_T = dev_list[p8,2]
    fv_x0 = dev_list[p8,3]
    fv_z = dev_list[p8,-1]
    k = np.argmin(fv_z)

    if debug:
        def get_nth_params(k):
            poresize = fv_x[k]
            N = fv_y[k]
            T = fv_T[k]
            x0 = fv_x0[k]
            print([k], "N, T, x0=", N, T, x0)
            return np.concatenate([[N, T, x0, me, mp, poresize], props])

        with plt.Dp():
            fig = plt.figure(figsize=(12, 5))
            fig.suptitle("guess_monopore_params_roughtly Debug & Proof", fontsize=20)
            ax1 = fig.add_subplot(121, projection="3d")
            ax2 = fig.add_subplot(122)
            ax1.set_title("fv vs. poresize vs. num_stays", fontsize=16)
            ax1.set_xlabel("poresize")
            ax1.set_ylabel("num_stays")
            ax1.set_zlabel("fv")
            ax1.plot(fv_x, fv_y, fv_z, "o")
            ax1.plot(fv_x[k], fv_y[k], fv_z[k], "o", color="red")
            # zmin, zmax = ax1.get_zlim()
            # ax.set_zlim(zmin, 1000)
            ax2.set_title("curves with selected params", fontsize=16)
            plot_params = get_nth_params(k)
            poresize, N = plot_params[[5,0]]
            rough_objective(plot_params[1:3], debug=True)
            ax2.plot(x, y, label="data")
            artists = plot_monopore_curves(ax2, x, plot_params, peak_rgs, return_artists=True)

            rax = ax1.inset_axes([1.05, 0.0, 0.3, 0.1])
            selecting = False
            check = CheckButtons(
                ax=rax,
                labels=["Selecting"],
                actives=[selecting],
                )
            def check_on_click(label):
                nonlocal selecting
                if label == "Selecting":
                    selecting = not selecting
                    fig.canvas.draw_idle()
            check.on_clicked(check_on_click)

            def on_click_to_select(event):
                if not selecting:
                    return
                if event.inaxes != ax1:
                    return

                ix, iy = event.xdata, event.ydata
                print("ix, iy=", ix, iy)
                p1, pane_idx = ax1._calc_coord(ix, iy)  # see matplotlib/lib/mpl_toolkits/mplot3d/axes3d.py as of matplotlib 3.8.3
                ix_, iy_, iz_ = p1
                # Find nearest point
                points = np.c_[fv_x, fv_y, fv_z]
                tree = cKDTree(points)
                dist, idx = tree.query([ix_, iy_, iz_])
                N_, poresize_, fv_ = fv_x[idx], fv_y[idx], fv_z[idx]
                print("idx=", idx, N_, poresize_, fv_)
                # Highlight nearest point
                ax1.scatter(fv_x[idx], fv_y[idx], fv_z[idx], c='r')     # task: should update the artist instead of adding new one

                plot_params = get_nth_params(idx)
                cy_list, ty = compute_monopore_curves(x, plot_params[0:6], peak_rgs, plot_params[6:])
                for k, y_ in enumerate(cy_list + [ty]):
                    artists[k].set_ydata(y_)            
                # plt.draw()
                fig.canvas.draw_idle()

            fig.canvas.mpl_connect('button_press_event', on_click_to_select)
            fig.tight_layout()
            # fig.subplots_adjust(right=0.7)
            ret = plt.show()
        if not ret:
            return

    cy_list = []
    for k, params in enumerate(peaks):
        cy = model(x, params)
        cy_list.append(cy)
    ty = np.sum(cy_list, axis=0)

    def plot_curves(ax, cy_list, ty, return_artists=False):
        ax.plot(x, y, label='data')
        artists = []
        for k, cy in enumerate(cy_list):
            curve, = ax.plot(x, cy, ":", label='component-%d' % k)
            artists.append(curve)
        curve, = ax.plot(x, ty, ":", color="red", label='model total')
        artists.append(curve)
        if return_artists:
            return artists

    # N, T, x0, me, mp, poresize
    poresize, N, T, x0 = dev_list[k,0:4]
    if debug:
        print("poresize, N, T, x0=", poresize, N, T, x0)
        print("peak_rgs=", peak_rgs)
        # print("rhov=", rhov)

    abort = False
    use_multi_factor_fv = False
    if use_multi_factor_fv:
        raw_moments = np.array([M[0] for M in egh_moments_list])
    optimize_poresize = False
    if not optimize_poresize:
        rhov = peak_rgs/poresize
        rhov[rhov > 1] = 1

    def scales_objective(p, title=None, return_curves=False, with_sliders=False):
        nonlocal abort, rhov
        if optimize_poresize:
            rp, N_, T_, x0_ = p[0:4]
            rhov = peak_rgs/rp
            rhov[rhov > 1] = 1
            scales_ = p[4:]
        else:
            N_, T_, x0_ = p[0:3]
            scales_ = p[3:]
        cy_list = []
        for k, (rho, scale) in enumerate(zip(rhov, scales_)):
            ni_ = N_*(1 - rho)**me
            ti_ = T_*(1 - rho)**mp
            cy = scale*robust_single_pore_pdf(x - x0_, ni_, ti_)
            cy_list.append(cy)
            # print([k],  x0_ + ni_*ti_ - m)    # m == x0_ + ni_*ti_, approximately
        ty_ = np.sum(cy_list, axis=0)
        if return_curves:
            return cy_list, ty_

        if use_multi_factor_fv:
            raw_moments_ = compute_raw_moments1(x, cy_list)
            props_ = compute_area_props(cy_list)

        if title is not None:
            figsize = (12, 5) if with_sliders else None
            print("params=", p)
            with plt.Dp():
                fig, ax = plt.subplots(figsize=figsize)
                ax.set_title(title)
                artists = plot_curves(ax, cy_list, ty_, return_artists=with_sliders)
                ax.legend()
                if with_sliders:
                    assert not optimize_poresize
                    update_params = p.copy()
                    slider_specs = [    ("N", 300, 4000, update_params[0]),
                                        ("T", 0, 3, update_params[1]),
                                        ("x0",  -500, 500, update_params[2]),
                                        ]
                    for k, scale in enumerate(scales_):
                        slider_specs.append(("Scale-%d" % k, 0, 2, update_params[3+k]))
                        
                    def slider_update(k, val):
                        print([k], "slider_update", val)
                        update_params[k] = val
                        redraw_params = np.concatenate([update_params[0:3], [me, mp, poresize], update_params[3:]])
                        cy_list, ty_ = compute_monopore_curves(x, redraw_params[0:6], peak_rgs, redraw_params[6:])
                        for k, (curve, y_) in enumerate(zip(artists, cy_list + [ty_])):
                            print([k], np.max(y_))
                            curve.set_ydata(y_)
                        fig.canvas.draw_idle()

                    slider_axes = []
                    sliders = []
                    for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
                        ax_ = fig.add_axes([0.8, 0.7 - 0.08*k, 0.11, 0.03])
                        slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
                        slider.on_changed(lambda val, k_=k: slider_update(k_, val))
                        slider_axes.append(ax_)
                        sliders.append(slider)                   

                    fig.tight_layout()
                    fig.subplots_adjust(right=0.6)
                else:
                    fig.tight_layout()
                ret = plt.show()
            if not ret:
                abort = True

        if use_multi_factor_fv:
            return np.sum((ty_ - ty)**2)**2 + 0.0001*np.sum((raw_moments_ - raw_moments)**2)**2 + 0.0001*np.sum((props_ - props)**2)**2
        else:
            return np.sum((ty_ - ty)**2)

    num_components = len(peak_rgs)

    init_x0 = x0
    column_params = [N, T, init_x0]
    adjust_params = column_params + [me, mp, poresize]
    adjusted_scales = adjust_scales_to_egh_params_array(x, y, adjust_params, peak_rgs, peaks)

    bounds = [(300, 5000), (0.01, 10), (init_x0-500, init_x0+500)] + [(0, 100)]*num_components
    if optimize_poresize:
        column_params.insert(0, poresize)
        bounds.insert(0, (20, MAX_PORESIZE))
    init_params = np.concatenate([column_params, adjusted_scales])
    if debug:
        scales_objective(init_params, title="before scales_objective minimize", with_sliders=True)
        if abort:
            return

    res = minimize(scales_objective, init_params, bounds=bounds, method="Nelder-Mead")
    if debug:
        scales_objective(res.x, title="after scales_objective minimize")
        if abort:
            return

        with plt.Dp(button_spec=["OK", "Cancel"]):
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(10,4))
            fig.suptitle("guess_monopore_params_roughtly")
            plot_curves(ax1, cy_list, ty)
            plot_curves(ax2, *scales_objective(res.x, return_curves=True))
            fig.tight_layout()
            ret = plt.show()
        if not ret:
            return

    if optimize_poresize:
        poresize, N, T, x0 = res.x[0:4]
        scales = res.x[4:]
    else:
        N, T, x0 = res.x[0:3]
        scales = res.x[3:]

    return np.concatenate([[N, T, x0, me, mp, poresize], scales])


"""
    ATSAS/Almerge.py

    Copyright (c) 2016-2023, SAXS Team, KEK-PF
"""
import sys
import os
import glob
import re
import numpy                as np
import subprocess
from OurSubprocess          import exec_subprocess
from molass_legacy._MOLASS.SerialSettings         import get_setting
from Result                 import Result
from molass_legacy.ATSAS.AutoRg import autorg_exe_array

dataline_re = re.compile( r'^\s*\d' )

# generator to read the result file
def read_data_lines( fh ):
    while True:
        try:
            line = fh.readline()
        except:
            # ignore bad data in the tail comment lines from almerge bug
            # UnicodeDecodeError: 'cp932' codec can't decode byte 0xe7 in position 5699: illegal multibyte sequence
            continue

        if not line:
            break
        if dataline_re.match( line ):
            yield line
        else:
            continue

def atsas_file_load( out_file ):
    fh = open( out_file )
    exz_array = np.loadtxt( read_data_lines(fh) )
    fh.close()
    return exz_array

class AlmergeExecutor:
    def __init__( self ):
        self.exe_path   = None
        if len( autorg_exe_array ) > 0:
            almerge_path = autorg_exe_array[0].replace( 'autorg.exe', 'almerge.exe' )
            if os.path.exists( almerge_path ):
                self.exe_path   = almerge_path

        self.out_line_re    = re.compile( r'(\d+)\s*-\s*(\d+)' )

    def execute( self, c_vector, datafiles, indeces, out_file ):
        if self.exe_path is None:
            return None

        self.datafiles  = datafiles
        self.c_vector   = c_vector

        # print( 'indeces=', indeces )
        # print( 'c_vector=', self.c_vector[indeces] )
        cmd = [ self.exe_path ]
        for i in indeces:
            cmd += [ '-c', '%g' % self.c_vector[i], self.datafiles[i] ]

        cmd += [ '-z', '-o', out_file ]

        out, err = exec_subprocess( cmd, shell=False )
        out = out.replace( r'\\', r'/' )

        print( out )
        if err != '':
            print( err )
            return None

        out_lines = out.split( '\n' )
        overlap_from_max = None
        for line in out_lines:
            m = self.out_line_re.search( line )
            if m:
                # print( m.group(1), m.group(2) )
                overlap_from = int( m.group(1) )
                if overlap_from_max is None or overlap_from > overlap_from_max:
                    overlap_from_max = overlap_from

        if not os.path.exists( out_file ):
            # TODO: error info
            return None

        exz_array = atsas_file_load( out_file )

        return Result( exz_array=exz_array, overlap_from_max=overlap_from_max )

    def execute_matrix(self, q, M, E, c_vector, out_file=None):
        from molass_legacy.KekLib.NumpyUtils import np_savetxt

        temp_folder = get_setting('temp_folder')
        datafiles = []
        for k, (data, error)in enumerate(zip(M.T, E.T)):
            file = temp_folder + '/almerge_in_%05d.dat' % k
            datafiles.append(file)
            np_savetxt(file, np.vstack([q, data, error]).T)

        indeces = np.arange(M.shape[1])
        if out_file is None:
            out_file = temp_folder + '/almerge_out.dat'

        return self.execute(c_vector, datafiles, indeces, out_file)



"""
    Optimizer.SimpleDebugUtils.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
from .FuncImporter import import_objective_function

def copycreate_optimizer(optimizer):
    class_code = optimizer.__class__.__name__
    optimizer_class = import_objective_function(class_code)
    new_optimizer = optimizer_class(
        optimizer.dsets,
        optimizer.n_components,
        uv_base_curve=optimizer.uv_base_curve,
        xr_base_curve=optimizer.xr_base_curve,
        qvector=optimizer.qvector,
        wvector=optimizer.wvector,
        composite=optimizer.composite,
        )
    return new_optimizer

def debug_optimizer(optimizer, debug_params=None, plot=False):
    new_optimizer = copycreate_optimizer(optimizer)
    if debug_params is None:
        debug_params = optimizer.init_params
    new_optimizer.prepare_for_optimization(debug_params)
    new_optimizer.objective_func(debug_params, plot=plot)


"""
    DatGnom.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import os
import re
from OurSubprocess import exec_subprocess
from Result import Result
from .AutoRg import autorg_exe_array

DEBUG = False
if DEBUG:
    import logging

class DatgnomExecutor:
    def __init__( self, exe_index=0):
        if DEBUG:
            self.logger = logging.getLogger(__name__)

        self.exe_path   = None
        if len( autorg_exe_array ) > 0:
            # examining datgnom5.exe first comes from molass_legacy.ATSAS 2.7.1
            # TODO: control this after getting the ATSAS version
            for name in [ 'datgnom4.exe', 'datgnom.exe' ]:
                datgnom_path = autorg_exe_array[exe_index].replace( 'autorg.exe', name )
                if os.path.exists( datgnom_path ):
                    self.exe_path   = datgnom_path
                    break

        self.datgnom4_compat = self.exe_path.find("datgnom4") >= 0

    def execute( self, in_file_, rg, out_file_, debug=True ):
        in_file = in_file_.replace('/', '\\')
        out_file = out_file_.replace('/', '\\')
        if self.exe_path is None or rg is None:
            return Result( Qrange=(None, None), Dmax=None, RgPr=None, RgPr_Error=None, IzPr=None, IzPr_Error=None )

        if debug:
            print("self.exe_path=", self.exe_path)

        cmd = [self.exe_path, '-r', '%g' % rg, '-o', out_file, in_file]
        if DEBUG:
            self.logger.info("cmd=%s", str(cmd))
        out, err = exec_subprocess( cmd, shell=False )
        # out = out.replace( '\\', '/' )

        if err != '':
            # print( err )
            # return None
            # atsas-2.8.2 produces warnings
            # TODO: check the exit code instead
            pass

        qrange      = None
        dmax        = None
        rg_pr       = None
        rg_pr_err   = None
        iz_pr       = None
        iz_pr_err   = None

        if self.datgnom4_compat:
            qrange_re = re.compile(r"Angular\s+range\s*:\s+from\s+(\S+)\s+to\s+(\S+)")
            dmax_re = re.compile( r'Dmax\s*=\s+(\d+\.?\d*)')
            rg_re   = re.compile( r'Rg\s+=\s+(\d+\.\S+)\s+\+-\s+(\d+\.\S+)' )
            iz_re   = re.compile( r'I\(0\)\s+=\s+(\d+\.\S+)\s+\+-\s+(\d+\.\S+)' )

            out_lines = out.split( '\n' )
            for line in out_lines:
                m = dmax_re.search( line )
                if m:
                    dmax = float(m.group(1))

            fh = open( out_file )
            for k, line in enumerate(fh):
                if line.find('Angular') >= 0:
                    m = qrange_re.search( line )
                    if m:
                        qrange = [float(m.group(i)) for i in (1, 2)]

                if line.find('Real space:') > 0:
                    # print([k], line)

                    m = rg_re.search( line )
                    if m:
                        rg_pr       = float(m.group(1))
                        rg_pr_err   = float(m.group(2))

                    m = iz_re.search( line )
                    if m:
                        iz_pr       = float(m.group(1))
                        iz_pr_err   = float(m.group(2))
                        break
            fh.close()
        else:
            qrange_re = re.compile(r"Angular\s+range:\s+(\S+)\s+to\s+(\S+)")
            dmax_re = re.compile(r"Real space range:\s+\S+\s+to\s+(\S+)")
            rg_re   = re.compile(r"Real space Rg:\s+(\S+)\s+\+-\s+(\S+)")
            iz_re   = re.compile(r"Real space I\(0\):\s+(\S+)\s+\+-\s+(\S+)")

            fh = open( out_file )
            for k, line in enumerate(fh):
                if line.find('Angular') >= 0:
                    m = qrange_re.search( line )
                    if m:
                        qrange = [float(m.group(i)) for i in (1, 2)]

                if line.find('Real space') > 0:
                    m = dmax_re.search( line )
                    if m:
                        dmax = float(m.group(1))

                    m = rg_re.search( line )
                    if m:
                        rg_pr       = float(m.group(1))
                        rg_pr_err   = float(m.group(2))

                    m = iz_re.search( line )
                    if m:
                        iz_pr       = float(m.group(1))
                        iz_pr_err   = float(m.group(2))
                        break
            fh.close()

        # print( dmax, rg_pr, rg_pr_err, iz_pr, iz_pr_err )

        return Result( Qrange=qrange, Dmax=dmax, RgPr=rg_pr, RgPr_Error=rg_pr_err, IzPr=iz_pr, IzPr_Error=iz_pr_err )

def positive_value( v ):
    v_ = float(v)
    return v_ if v_ > 0 else None

spaces_re = re.compile( r'\s+' )

def datgnom_read_data(datgnom_out_file, null_value=None, null_func=positive_value):

    expr_data_list = []
    real_space_list = []

    expr_data_found     = False
    real_space_found    = False
    fh = open( datgnom_out_file )
    for k, line in enumerate(fh):
        if expr_data_found:
            pass
        else:
            # if line.find( 'Experimental Data and Fit' ) > 0:
            if line.find( 'J EXP' ) > 0:
                expr_data_found = True
                # print( 'expr_data_found' )
            continue

        if real_space_found:
            values = re.split( spaces_re, line )
            if len(values) == 5 and values[1][0].isdigit():
                real_space_list.append( [float(v) for v in values[1:-1] ] )

        else:
            # if line.find( 'Real Space Data' ) > 0:
            if line.find( 'Distance distribution' ) > 0:
                real_space_found = True
                # print( 'real_space_found' )
                continue
            else:
                pass

            values = re.split( spaces_re, line )
            if len(values) == 4 and values[1][0].isdigit():
                # print( line, values )
                expr_data_list.append( [ float(values[1]) ] + [ null_value ]*3 + [ null_func(values[2]) ] )

            elif len(values) == 7 and values[1][0].isdigit():
                # print( line, values )
                expr_data_list.append( [ float(values[1]) ] + [ null_func(v) for v in values[2:-1] ] )

    fh.close()

    return expr_data_list, real_space_list



"""
    Selective.ModelFrameStatus.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

class ModelFrameStatus:
    def __init__(self):
        self.has_info = False

    def save_status(self, advanced_frame):
        self.has_info = True

    def restore_status(self, advanced_frame):
        if not self.has_info:
            return



# coding: utf-8
"""
    SmoothFactorizer.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
from scipy import optimize
from SmoothnessPenalty import SmoothnessPenalty
MAX_ITERATION_FMIN_CG   = 10000

class SmoothFactorizer:
    def __init__(self, M, E, c):
        self.M  = M
        self.E  = E
        self.c  = c

    def solve(self, with_const=False, apositive=False, smoothed=False):
        self.apositive = apositive
        self.smoothed = smoothed
        self.ignore_bq =False
        self.penalty_weights = [ 0.5, 0.01, 1.0, 0.1 ]
        c_ = self.c
        c_list = [c_, c_**2]
        if with_const:
            c_list.append( np.ones(len(c_)) )

        self.C = np.array( c_list )
        self.Cpinv = np.linalg.pinv(self.C)
        self.P_init = np.dot(self.M, self.Cpinv)
        # self.e = np.sqrt( np.dot(self.E**2, self.Cpinv**2) )
        # print( self.P.shape, self.e.shape )

        self.sp = SmoothnessPenalty(self.P_init.shape[0])

        self.Zero = np.zeros( self.P_init.shape )
        n = self.P_init.shape[1]
        self.P2A = np.zeros( (n, n) )
        for i in range(0, n, 2):
            self.P2A[i,i] = 1

        self.P = self.mimimize()

    def compute_error( self, P_flat ):
        P = P_flat.reshape( self.P_init.shape )
        reconstructed = np.dot( P, self.C )

        # compute a squared Frobenius norms
        rec_error = np.linalg.norm( reconstructed - self.M )**2

        if self.apositive:
            A   = np.dot( P, self.P2A )
            negA = np.min( [ self.Zero, A ], axis=0 )
            neg_error = np.linalg.norm( negA )**2
        else:
            neg_error = 0
        error   = rec_error + neg_error*self.penalty_weights[0]

        if self.smoothed:
            if self.ignore_bq:
                penalties = self.sp.get_penalties_bq_ignore(P_flat, self.penalty_weights[2])
                error += np.sum(penalties)
            else:
                penalties = self.sp.get_penalties(P_flat, self.penalty_weights[2:])
                error += np.sum(penalties)

        return error

    def compute_error_grad( self, P_flat ):
        P = P_flat.reshape( self.P_init.shape )
        reconstructed = np.dot( P, self.C )

        # compute the derivative of the squared Frobenius norm
        # || P@C - D ||**2 ==> 2*( P@C - D )@C.T
        rec_grad = 2 * np.dot( reconstructed - self.M, self.C.T )

        if self.apositive:
            A   = np.dot( P, self.P2A )
            negA = np.min( [ self.Zero, A ], axis=0 )
            neg_grad = 2 * negA
            neg_grad_ = neg_grad.flatten()*self.penalty_weights[0]
        else:
            neg_grad_ = 0

        error_grad  = rec_grad.flatten() + neg_grad_

        if self.smoothed:
            if self.ignore_bq:
                error_grad += self.sp.get_penalty_diff_bq_ignore(P_flat, self.penalty_weights[2])
            else:
                error_grad += self.sp.get_penalty_diff(P_flat, self.penalty_weights[2:])

        return error_grad

    def mimimize( self ):
        ret = optimize.fmin_cg( self.compute_error, self.P_init.flatten(), fprime=self.compute_error_grad,
                maxiter=MAX_ITERATION_FMIN_CG,
                full_output=True )
        P_flat = ret[0]
        return P_flat.reshape(self.P_init.shape)



# coding: utf-8
"""

    SubProcess.py

    Copyright (c) 2020, Masatsuyo Takahashi, KEK-PF

"""

import subprocess

"""
    thanks to https://github.com/SublimeText/VintageEx/blob/master/plat/windows.py
"""
def get_startup_info():
    # Hide the child process window.
    startupinfo = subprocess.STARTUPINFO()
    startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
    return startupinfo

class Popen(subprocess.Popen):
    def __init__(self, cmd, **kwargs):
        subprocess.Popen.__init__(self, cmd, startupinfo=get_startup_info(), **kwargs)



"""
    QuickAnalysis.ModeledPeaks.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.interpolate import UnivariateSpline
from scipy.optimize import fsolve, minimize
from scipy.special import erf
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.ElutionCurveModels import egh, egha
from molass_legacy.Peaks.ElutionModels import compute_moments, compute_egh_params
from molass_legacy._MOLASS.SerialSettings import get_setting

RESID_RATIO_LIMIT = 0.03
H_RATIO_LIMIT = 0.03
PENALTY_SCALE = 1e3
TAU_BOUND_RATIO = get_setting("TAU_BOUND_RATIO")    # tau <= sigma*TAU_BOUND_RATIO
MAPPING_ADJUST_ALLOW_RATIO = 0.1
DEFAULT_MIN_AREA_PROP = 0.02

def gaussian(x, h, m, s):
    return h * np.exp(-((x-m)/s)**2)

def gaussian_integral(x, h, m, s):
    return np.sqrt(np.pi) * h * s * erf((x-m)/s) / 2

def get_a_peak(x, y, refine=True, model=None, affine=False, debug=False):
    if model is None:
        from molass_legacy.Models.ElutionCurveModels import EGH, EGHA
        model = EGHA() if affine else EGH()
    model_func = model.func     # task: remove model_func

    pt = np.argmax(y)
    mu_init = x[pt]
    max_y = y[pt]
    smax = len(x)//2
    s_init = smax/2

    for s in range(3, smax):
        slice_ = slice(max(0, pt-s), min(len(x), pt+s))
        x_ = x[slice_]
        y_ = y[slice_]
        gy = gaussian(x_, max_y, mu_init, s)
        if False:
            print("slice_=", slice_)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("%s plot" % str([s]))
                ax.plot(x, y)
                ax.plot(x_, gy, ":")
                plt.show()
        negative_ratio = len(np.where(y_ - gy < 0)[0])/len(x_)
        if debug:
            print([s], "negative_ratio=", negative_ratio)
        if negative_ratio > 0.5:
            s_init = s
            break

    w = np.where(y[0:pt] < 0)[0]
    fx_lim = x[0] if len(w) == 0 else x[w[-1]]
    w = np.where(y[pt:] < 0)[0]
    tx_lim = x[-1] if len(w) == 0 else x[pt + w[0]]
    max_width = tx_lim - fx_lim

    bnds = ((0, max_y*1.5), (x[0], x[-1]), (0, max_width), (0, max_width))
    y_for_opt = y.copy()
    width = int(s_init*2)
    left = max(0, pt-width)
    right = min(len(x), pt+width)
    y_for_opt[0:left] = 0
    y_for_opt[right:] = 0
    y_for_opt[y_for_opt < 0] = 0

    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("y_for_opt proof")
            ax.plot(x, y)
            ax.plot(x, y_for_opt, ":")
            plt.show()

    M = compute_moments(x, y_for_opt)
    init_params = mu_init, s_init, 0
    tR, sigma, tau = compute_egh_params(init_params, M)

    h_init = y[pt]

    if affine:
        a_init = s_init * 0.5
        affine_bnds = bnds + ((-a_init, a_init),)

        def obective_affine(p):
            h, mu, sigma, tau, a = p
            my = model_func(x, h, mu, sigma, tau, a)
            return np.sum((my - y_for_opt)**2) + PENALTY_SCALE*(min(0, tau)**2 + min(0, sigma*TAU_BOUND_RATIO - tau)**2)

        ret = minimize(obective_affine, (h_init, tR, sigma, tau, 0), bounds=affine_bnds)
    else:
        def obective(p):
            h, mu, sigma, tau = p
            my = model_func(x, h, mu, sigma, tau)
            return np.sum((my - y_for_opt)**2) + PENALTY_SCALE*(min(0, tau)**2 + min(0, sigma*TAU_BOUND_RATIO - tau)**2)

        # cons = ({'type': 'ineq', 'fun': lambda x:  x[2] - x[3] },)
        # it seems better to have this included in the obective func
        # than to add constraints=cons below
        ret = minimize(obective, (h_init, tR, sigma, tau), bounds=bnds)

    if debug:
        print("s_init=", s_init)
        print("sigma=", sigma)
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("get_a_peak: s_init, opt_sinit")
            ax1.plot(x, y)
            ax1.plot(x, gaussian(x, max_y, mu_init, s_init), label="gaussian(..., s_init)")
            ax1.plot(x, gaussian(x, max_y, mu_init, sigma), label="gaussian(..., sigma)")
            ax2.plot(x, y)
            ax2.plot(x, y_for_opt, ":")
            ax2.plot(x, model_func(x, h_init, tR, sigma, tau), label="egh(..., sigma, ...)")
            ax2.plot(x, model_func(x, *ret.x), label="egh(optimized)")
            for ax in ax1, ax2:
                ax.legend()
            fig.tight_layout()
            plt.show()

    return ret.x

def recognize_peaks(x, y, num_peaks=5, exact_num_peaks=None, affine=False, model=None, min_area_prop=None, correct=True, debug=False):
    """
    moved here due to 
    ImportError: cannot import name 'get_corrected' from partially initialized module 'LPM' (most likely due to a circular import) (.../DataStructure/LPM.py)
    """
    from molass_legacy.DataStructure.LPM import get_corrected

    if model is None:
        from molass_legacy.Models.ElutionCurveModels import EGH, EGHA
        if affine:
            model = EGHA()
        else:
            model = EGH()

    decide_num_peaks = False
    if min_area_prop is None:
        min_area_prop = DEFAULT_MIN_AREA_PROP
    else:
        decide_num_peaks = True

    get_exact_num_peaks = exact_num_peaks is not None
    max_num_peaks = exact_num_peaks if get_exact_num_peaks else num_peaks

    if max_num_peaks is None:
        max_num_peaks = 7   # MAX_NUM_PEAKS

    if debug:
        print("recognize_peaks: max_num_peaks=", max_num_peaks, "decide_num_peaks=", decide_num_peaks)

    y_max = np.max(y)

    if correct:
        y_copy = get_corrected(y, x=x)
    else:
        y_copy = y.copy()

    peaks_list = []
    if decide_num_peaks:
        total_area = np.sum(y_copy)
    for k in range(max_num_peaks):
        params = get_a_peak(x, y_copy, refine=k == 0, model=model, affine=affine, debug=debug)
        if decide_num_peaks:
            y_model = model(x, params)
            area = np.sum(y_model)
            area_ratio = area/total_area
            if area_ratio < min_area_prop:
                break
        else:
            h_ratio = params[0]/y_max
            if h_ratio < H_RATIO_LIMIT:
                print([k], "h_ratio=", h_ratio)
                if not get_exact_num_peaks:
                    break
            y_model = model(x, params)
        y_resid = y_copy.copy()
        y_resid -= y_model

        resid_ratio = np.max(y_resid)/y_max
        if debug:
            print("resid_ratio=", resid_ratio)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("recognize_peaks with num_peaks=%d" % num_peaks)
                ax.plot(x, y)
                ax.plot(x, y_model)
                ax.plot(x, y_resid)
                plt.show()

        peaks_list.append(params)

        if get_exact_num_peaks:
            pass
        else:
            if resid_ratio < RESID_RATIO_LIMIT:
                break

        y_copy = y_resid

    if decide_num_peaks:
        ret_peaks_list = []
        props = model.get_proportions(x, peaks_list)
        for i in np.where(props > min_area_prop)[0]:
            ret_peaks_list.append(peaks_list[i])
    else:
        ret_peaks_list = peaks_list

    sorted_list = sorted(ret_peaks_list, key=lambda x: x[1])

    return sorted_list

def plot_curve(ax, x, y, peaks, color='C0', model=None, baseline=None, dcurves=None, legend=True, return_markers=False):
    if model is None:
        from molass_legacy.Models.ElutionCurveModels import EGH, EGHA
        if len(peaks[0]) == 5:
            model = EGHA()
        else:
            model = EGH()

    ax.plot(x, y, color=color, label='data', lw=3)
    ty = np.zeros(len(y))
    k = 0
    markers = []
    for param in peaks:
        cy = model(x, param)
        ty += cy
        cy_ = cy if baseline is None else cy + baseline
        if dcurves is None:
            ax.plot(x, cy_, ':', label='component-%d' % (k+1))
        else:
            dcurves.add( x, cy_, ":", label='component-%d' % (k+1))
        px, py = model.get_peaktop_xy(x, param)
        if px > x[0] and px < x[-1] and py > 0:
            marker, = ax.plot(px, py, "o", color="yellow")
            markers.append(marker)

        k += 1
    ty_ = ty if baseline is None else ty + baseline
    ax.plot(x, ty_, ':', color='red', label='component total', lw=3)
    if baseline is not None:
        ax.plot(x, baseline, color='red', label='baseline')

    if legend:
        ax.legend()

    if return_markers:
        return markers

    return ty

def adjust_peak_heights(a, b, xr_x, xr_y, peaks, uv_x, uv_y, affine=False, debug=False):
    uv_allow = len(uv_x)*MAPPING_ADJUST_ALLOW_RATIO

    def objective(p):
        ty = np.zeros(len(uv_y))
        k = 0
        # range_penalty is required in cases like 20201007_2
        # which might be better treated in trimming
        range_penalty = max(0, abs(xr_x[0]*a + b - uv_x[0]) - uv_allow)**2 + max(0, abs(xr_x[-1]*a + b - uv_x[-1]) - uv_allow)**2
        for params in peaks:
            h, mu, sigma, tau = params[0:4]
            h_ = p[k]
            mu_ = a*mu + b
            sigma_ = a*sigma
            tau_ = a*tau
            if affine:
                ty += egha(uv_x, h_, mu_, sigma_, tau_, params[4])
            else:
                ty += egh(uv_x, h_, mu_, sigma_, tau_)
            k += 1
        return np.sum((ty - uv_y)**2) + range_penalty*PENALTY_SCALE

    h_init = []
    uv_spline = UnivariateSpline(uv_x, uv_y, s=0, ext=3)    # exit=3 would be safe for end points

    if debug:
        debug_peaks = []
    init_heights = []
    for params in peaks:
        h, mu, sigma, tau = params[0:4]
        init_heights.append(h)
        mu_ = a*mu + b
        h_ = uv_spline(mu_)
        h_init.append(h_)
        if debug:
            sigma_ = a*sigma
            tau_ = a*tau
            if affine:
                debug_peaks.append((h_, mu_, sigma_, tau_, params[4]))
            else:
                debug_peaks.append((h_, mu_, sigma_, tau_))

    if debug:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("adjust_peak_heights initial state")
            plot_curve(ax1, uv_x, uv_y, debug_peaks)
            plot_curve(ax2, xr_x, xr_y, peaks, color="orange")
            fig.tight_layout()
            plt.show()

    min_h = max(0, min(np.min(init_heights), np.min(h_init)))
    bounds = [(min_h, 100)] * len(peaks)
    # print("min_h=", min_h, "h_init=", h_init)

    ret = minimize(objective, h_init, method="Nelder-Mead", bounds=bounds)

    if debug:
        a_ = 1/a
        b_ = - b/a
        mp_x = uv_x*a_ + b_
        ratio = np.max(xr_y)/np.max(uv_y)
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("adjust_peak_heights optimized state")
            plot_curve(ax1, uv_x, uv_y, debug_peaks)
            plot_curve(ax2, xr_x, xr_y, peaks, color="orange", legend=False)
            ax2.plot(mp_x, uv_y*ratio, label="mapped")
            ax2.legend()
            fig.tight_layout()
            plt.show()

    ret_peaks = []
    k = 0
    for params in peaks:
        h, mu, sigma, tau = params[0:4]
        h_ = ret.x[k]
        mu_ = a*mu + b
        sigma_ = a*sigma
        tau_ = a*tau
        if affine:
            ret_peaks.append((h_, mu_, sigma_, tau_, params[4]))
        else:
            ret_peaks.append((h_, mu_, sigma_, tau_))
        k += 1

    return ret_peaks

def get_curve_xy_impl(sd, baseline_type=1, return_details=False, debug=False):
        if debug:
            print("get_curve_xy_impl: baseline_type=", baseline_type)

        baselines = None
        baseline_params = None

        X, E, qv, xr_curve = sd.get_xr_data_separate_ly()
        U, _, wv, uv_curve = sd.get_uv_data_separate_ly()
        xr_x = xr_curve.x
        xr_y = xr_curve.y
        uv_x = uv_curve.x
        uv_y = uv_curve.y

        if return_details:
            from molass_legacy.KekLib.BasicUtils import Struct
            from molass_legacy.Baseline.Constants import SLOPE_SCALE

            if baseline_type == 0:
                assert False
            else:
                if baseline_type == 1:
                    if debug:
                        from importlib import reload
                        import Baseline.LinearBaseline
                        reload(Baseline.LinearBaseline)
                        import UV.UvPreRecog
                        reload(UV.UvPreRecog)
                    from molass_legacy.Baseline.LinearBaseline import LinearBaseline as XrBaselineClass
                    from molass_legacy.UV.UvPreRecog import UvPreRecog
                elif baseline_type == 2:
                    if debug:
                        from importlib import reload
                        import Baseline.IntegralBaseline
                        reload(Baseline.IntegralBaseline)            
                    from molass_legacy.Baseline.IntegralBaseline import IntegralBaseline as XrBaselineClass
                elif baseline_type == 3:
                    if debug:
                        from importlib import reload
                        import Baseline.FoulingBaseline
                        reload(Baseline.FoulingBaseline)            
                    from molass_legacy.Baseline.FoulingBaseline import FoulingBaseline as XrBaselineClass
                else:
                    assert False

                baseline_objects = []
                baselines = []
                baseline_params = []

                pre_recog = sd.pre_recog
                if pre_recog is None:
                    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
                    pre_recog = PreliminaryRecognition(sd)

                upr = UvPreRecog(sd, pre_recog, debug=False)
                base_curve = upr.base_curve
                baseline_objects.append(base_curve)
                ty = None   # dummy
                params = upr.init_params.copy()
                params[4:6] /= SLOPE_SCALE
                baselines.append(base_curve(uv_x, params, ty))
                baseline_params.append(params)

                # for x, y in [(uv_x, uv_y), (xr_x, xr_y)]:
                for x, y in [(xr_x, xr_y)]:
                    baseline = XrBaselineClass(x, y)
                    baseline_objects.append(baseline)
                    baselines.append(baseline.yb)
                    baseline_params.append(baseline.params)
                # uv_y = uv_y - baselines[0]
                # xr_y = xr_y - baselines[1]

            details = Struct(
                        uv_curve=uv_curve,
                        xr_curve=xr_curve,
                        baseline_objects=baseline_objects,
                        baselines=baselines,
                        baseline_params=baseline_params
                        )

            if debug:
                with plt.Dp():
                    fig, axes = plt.subplots(ncols=2, figsize=(12,5))
                    fig.suptitle("get_curve_xy_impl return_details")
                    for ax, (x, y), bl in zip(axes, [(uv_x, uv_y), (xr_x, xr_y)], baselines):
                        ax.plot(x, y)
                        if bl is not None:
                            ax.plot(x, bl, color="red")
                    fig.tight_layout()
                    plt.show()

            return uv_x, uv_y, xr_x, xr_y, details
        else:
            return uv_x, uv_y, xr_x, xr_y

def get_modeled_peaks_impl(a, b, uv_x, uv_y, xr_x, xr_y, num_peaks,
                           peaks=None,
                           exact_num_peaks=None,
                           affine=False, min_area_prop=None, debug=False):
    if debug:
        print("uv_y=", uv_y)
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            ax1.set_title("UV")
            ax2.set_title("XR")
            ax1.plot(uv_x, uv_y)
            ax2.plot(xr_x, xr_y)
            fig.tight_layout()
            plt.show()

    if peaks is None:
        peaks = recognize_peaks(xr_x, xr_y, num_peaks=num_peaks, exact_num_peaks=exact_num_peaks,
                            affine=affine, min_area_prop=min_area_prop, debug=debug)

    uv_peaks = adjust_peak_heights(a, b, xr_x, xr_y, peaks, uv_x, uv_y, affine=affine, debug=debug)
    xr_peaks = adjust_peak_heights(1, 0, xr_x, xr_y, peaks, xr_x, xr_y, affine=affine)

    if debug:
        from .ModeledPeaksTester import plot_modeled_peaks
        plot_modeled_peaks(uv_x, uv_y, xr_x, xr_y, uv_peaks, xr_peaks, a, b, suptitle="get_modeled_peaks_impl: return")

    return np.asarray(uv_peaks), np.asarray(xr_peaks)

def get_proportions(x, peaks):
    areas = []
    for h, mu, sigma, tau in peaks:
        y = egh(x, h, mu, sigma, tau)
        areas.append(np.sum(y))
    return np.array(areas)/np.sum(areas)

def get_peak_region_ends(x, peaks):
    ends = []
    for j in [0, -1]:
        h, mu, sigma, tau = peaks[j]
        sign = -1 if j == 0 else +1
        pos = int(mu + sign*3*(sigma+tau))
        ends.append(min(max(0, pos), len(x)))
    return ends

"""
    this class is experimantal
"""
class PeakRegionLrf:
    def __init__(self, D, E, x, peaks):
        f, t = get_peak_region_ends(x, peaks)
        x_ = x[f:t]
        y_list = []
        for p in peaks:
            y_list.append(egh(x_, *p))
        C = np.array(y_list)
        D_ = D[:,f:t]
        P = D_ @ np.linalg.pinv(C)

        E_ = E[:,f:t]
        D_pinv = np.linalg.pinv(D_)
        W = np.dot(D_pinv, P)
        Pe = np.sqrt(E_**2 @ W**2)

        self.lrf_results = P, Pe
        self.ends = f,t

    def compute_rgs(self, qv):
        from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
        P, Pe = self.lrf_results
        ret_rgs = []
        for y, ye in zip(P.T, Pe.T):
            sg = SimpleGuinier(np.array([qv, y, ye]).T)
            ret_rgs.append(sg.Rg)
        return ret_rgs



"""
    Guinier/RgComputer.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from molass_legacy.DataStructure.AnalysisRangeInfo import shift_range_from_to_by_x

def compute_rgs(M, E, qv, C, paired_ranges, x=None, return_trs=False, return_qualities=False, debug=False):
    assert C.shape[0] == len(paired_ranges)

    ret_rgs = []
    ret_trs = []
    peak_rgs = []
    peak_trs = []
    qualities = []
    includes_nan = False
    for k, prange in enumerate(paired_ranges):
        target_i = k
        temp_rgs = []
        temp_trs = []
        temp_qualities = []
        for f, t in prange.get_fromto_list():
            tr_ = (f+t)/2
            ret_trs.append(tr_)
            if x is not None:
                f, t = shift_range_from_to_by_x(x, f, t)
            if debug:
                print([k], f, t)
            slice_ = slice(f,t+1)
            M_ = M[:,slice_]
            C_ = C[:,slice_]
            E_ = E[:,slice_]
            P_ = M_ @ np.linalg.pinv(C_)
            Minv = np.linalg.pinv(M_)
            W = Minv @ P_
            Pe = np.sqrt(E_**2 @ W**2)
            data = np.array([qv, P_[:,target_i], Pe[:,target_i]]).T
            sg = SimpleGuinier(data)
            if sg.Rg is None or sg.Rg < 10:
                rg = np.nan
                includes_nan = True
            else:
                rg = sg.Rg
            ret_rgs.append(rg)
            temp_rgs.append(rg)
            temp_trs.append(tr_)
            temp_qualities.append(sg.basic_quality)
        peak_rgs.append(np.nanmean(temp_rgs))
        peak_trs.append(np.nanmean(temp_trs))
        qualities.append(np.nanmean(temp_qualities))

    if includes_nan:
        peak_rgs, qualities = interpolate_nan(peak_rgs, peak_trs, qualities)

    if return_trs:
        if return_qualities:
            return np.array(ret_rgs), np.array(ret_trs), np.array(peak_rgs), np.array(peak_trs), np.array(qualities)
        else:
            return np.array(ret_rgs), np.array(ret_trs), np.array(peak_rgs), np.array(peak_trs)
    else:
        return np.array(ret_rgs)

def interpolate_nan(peak_rgs, peak_trs, qualities, debug=False):
    """
    task: make this more reliable
    """
    isnan_indeces = np.where(np.isnan(peak_rgs))[0]
    if len(isnan_indeces) == 0:
        return peak_rgs, qualities

    # as in 20161104/BL-6A/pH6
    import logging
    from scipy.optimize import minimize
    from scipy.interpolate import UnivariateSpline
    logger = logging.getLogger(__name__)
    logger.info("interpolate_nan: isnan_indeces=%s", isnan_indeces)
    temp_rgs = np.asarray(peak_rgs).copy()
    temp_trs = np.asarray(peak_trs)
    temp_qualities = qualities.copy()
    valid_indeces = np.where(np.logical_not(np.isnan(temp_rgs)))[0]
    x = temp_trs[valid_indeces]
    y = temp_rgs[valid_indeces]
    """
    tr = t0 + NT*(1 - rg/poresize)**3
    """
    def objective(p):
        t0, NT, poresize = p
        rhov = x/poresize
        rhov[rhov > 1] = 1
        x_ = t0 + NT*(1 - rhov)**3
        return np.sum((x_ - x)**2)

    t0_init = np.min(x) - 30
    res = minimize(objective, (t0_init, 300, 80))
    t0, NT, poresize = res.x
    y_ = np.linspace(70, 30, 20)
    rhov = y_/poresize
    rhov[rhov > 1] = 1
    x_ = t0 + NT*(1 - rhov)**3
    spline = UnivariateSpline(x_, y_)

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        logger.info("interpolate_nan: x=%s, y=%s", x, y)

        fig, ax = plt.subplots()
        ax.set_title("interpolate_nan debug")
        ax.plot(x, y, 'o')
        ax.plot(x_, y_, ":")
        ax.plot(x_, spline(x_), ":")
        t_ = temp_trs[isnan_indeces]
        ax.plot(t_, spline(t_), "o", color="red")
        fig.tight_layout()
        plt.show()

    for i in isnan_indeces:
        rg = spline(peak_trs[i])
        temp_rgs[i] = rg
        temp_qualities[i] = 0
        logger.info("interpolate_nan: %d-th Rg has been interpolated as rg=%.3g using peak_trs[%d]=%.3g ", i, rg, i, temp_trs[i])
    return temp_rgs, temp_qualities


# coding: utf-8
"""
    MatsukenGmm.py.

    adapted from:
        EMアルゴリズム徹底解説
        https://qiita.com/kenmatsu4/items/59ea3e5dfa3d4c161efb

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import os
import numpy as np
import numpy.random as rd
import scipy as sp
from scipy import stats as st
from collections import Counter

import matplotlib
from matplotlib import font_manager
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib import rc
import matplotlib.animation as ani
plt.rcParams["patch.force_edgecolor"] = True
#rc('text', usetex=True)
import seaborn as sns
sns.set(style="whitegrid", palette="muted", color_codes=True)
sns.set_style("whitegrid", {'grid.linestyle': '--'})
red = sns.xkcd_rgb["light red"]
green = sns.xkcd_rgb["medium green"]
blue = sns.xkcd_rgb["denim blue"]

c = ['r', 'g', 'b']
n = [200, 150, 150]
N = np.sum(n)
D = 2
K = 3

def calc_likelihood(data, mu, sigma, pi, K):
    likelihood = np.zeros((np.sum(n), 3))
    for k in range(K):
        likelihood[:, k] = [pi[k]*st.multivariate_normal.pdf(d, mu[k], sigma[k]) for d in data]
    return likelihood

def calc_prob_gmm(data, mu, sigma, pi, K):
    return [[pi[k]*st.multivariate_normal.pdf(d, mu[k], sigma[k]) for k in range(K)] for d in data]

def print_gmm_contour(mu, sigma, pi, K):
    # display predicted scores by the model as a contour plot
    X, Y = np.meshgrid(np.linspace(min_x, max_x), np.linspace(min_y, max_y))
    XX = np.array([X.ravel(), Y.ravel()]).T
    Z = np.sum(np.asanyarray(calc_prob_gmm(XX, mu, sigma, pi, K)), axis=1)
    Z = Z.reshape(X.shape)
    CS = plt.contour(X, Y, Z, alpha=0.2, zorder=-100)
    
    plt.title('pdf contour of a GMM')

class GMM:
    def __init__(self, k, max_iter):
        self.k = k
        self.max_iter = max_iter

    def fit(self, X):
        pass

def generate_demo_data():
    global data

    seed = 77

    mu_true = np.asanyarray(
         [[0.2, 0.5],
          [1.2, 0.5],
          [2.0, 0.5]])

    D = mu_true.shape[1]

    sigma_true = np.asanyarray(
            [ [[0.1,  0.085],[ 0.085, 0.1]],
              [[0.1, -0.085],[-0.085, 0.1]],
              [[0.1,  0.085],[ 0.085, 0.1]]
            ])

    rd.seed(seed)
    org_data = None #np.empty((np.sum(n), 3))
    for i in range(3):
        print("check: ", i, mu_true[i], sigma_true[i], np.linalg.det(sigma_true[i]))
        #org_data.append(st.multivariate_normal.rvs(mean=mu[i], cov=sigma[i], size=n[i]))
        if org_data is None:
            org_data = np.c_[st.multivariate_normal.rvs(mean=mu_true[i], cov=sigma_true[i], size=n[i]), np.ones(n[i])*i]
        else:
            org_data = np.r_[org_data, np.c_[st.multivariate_normal.rvs(mean=mu_true[i], 
                                                                        cov=sigma_true[i], 
                                                                        size=n[i]), np.ones(n[i])*i]]
            
    # plot generated data        

    plt.figure(figsize=(12, 5))
    for i in range(3):
        plt.scatter(org_data[org_data[:,2]==i][:,0], org_data[org_data[:,2]==i][:,1], s=30, c=c[i], alpha=0.5)
        
    # drop true cluster label
    data = org_data[:,0:2].copy()
    return data

def initilize():
    global mu, sigma, pi
    global max_x, min_x, max_y, min_y

    # initialize pi
    pi = np.zeros(K)
    for k in range(K):
        if k == K-1:
            pi[k] = 1 - np.sum(pi)
        else:
            pi[k] = 1/K
    print('init pi:', pi)

    # initialize mu
    max_x, min_x = np.max(data[:,0]), np.min(data[:,0])
    max_y, min_y = np.max(data[:,1]), np.min(data[:,1])
    mu = np.c_[rd.uniform(low=min_x, high=max_x, size=K), rd.uniform(low=min_y, high=max_y, size=K) ]
    print('init mu:\n', mu)

    # visualize for check
    #plt.figure(figsize=(12,8))
    plt.figure(figsize=(12, 5))
    plt.scatter(data[:,0], data[:,1], s=30, c='gray', alpha=0.5, marker="+")

    for i in range(3):
        plt.scatter([mu[i, 0]], [mu[i, 1]], c=c[i], marker='o')
        
    plt.show()

    # initialize sigma
    sigma = np.asanyarray(
            [ [[0.1,  0],[ 0, 0.1]],
              [[0.1,  0],[ 0, 0.1]],
              [[0.1,  0],[ 0, 0.1]] ])

    # calculate likelihood
    likelihood = calc_likelihood(data, mu, sigma, pi, K)
    print('initial sum of log likelihood:', np.sum(np.log(likelihood)))

    print('pi:\n', pi)
    print('mu:\n', mu)
    print('sigma:\n', sigma)

    plt.title('initial state')

def animate(nframe):
    global mu, sigma, pi

    print('nframe:', nframe)
    plt.clf()
    
    if nframe <= 3:
        print('initial state')
        plt.scatter(data[:,0], data[:,1], s=30, c='gray', alpha=0.5, marker="+")
        for i in range(3):
            plt.scatter([mu[i, 0]], [mu[i, 1]], c=c[i], marker='o', edgecolors='k', linewidths=1)
        print_gmm_contour(mu, sigma, pi, K)
        plt.title('initial state')
        return

    # E step ========================================================================
    # calculate responsibility(負担率)
    likelihood = calc_likelihood(data, mu, sigma, pi, K)
    #gamma = np.apply_along_axis(lambda x: [xx/np.sum(x) for xx in x] , 1, likelihood)
    gamma = (likelihood.T/np.sum(likelihood, axis=1)).T
    N_k = [np.sum(gamma[:,k]) for k in range(K)]

    # M step ========================================================================

    # caluculate pi
    pi =  N_k/N

    # calculate mu
    tmp_mu = np.zeros((K, D))

    for k in range(K):
        for i in range(len(data)):
            tmp_mu[k] += gamma[i, k]*data[i]
        tmp_mu[k] = tmp_mu[k]/N_k[k]
        #print('updated mu[{}]:\n'.format(k) , tmp_mu[k])
    mu_prev = mu.copy()
    mu = tmp_mu.copy()
    #print('updated mu:\n', mu)

    # calculate sigma
    tmp_sigma = np.zeros((K, D, D))

    for k in range(K):
        tmp_sigma[k] = np.zeros((D, D))
        for i in range(N):
            tmp = np.asanyarray(data[i]-mu[k])[:,np.newaxis]
            tmp_sigma[k] += gamma[i, k]*np.dot(tmp, tmp.T)
        tmp_sigma[k] = tmp_sigma[k]/N_k[k]

        #print('updated sigma[{}]:\n'.format(k) , tmp_sigma[k])
    sigma = tmp_sigma.copy()

    # calculate likelihood
    prev_likelihood = likelihood
    likelihood = calc_likelihood(data, mu, sigma, pi, K)
    
    prev_sum_log_likelihood = np.sum(np.log(prev_likelihood))
    sum_log_likelihood = np.sum(np.log(likelihood))
    diff = prev_sum_log_likelihood - sum_log_likelihood
    
    print('sum of log likelihood:', sum_log_likelihood)
    print('diff:', diff)

    print('pi:', pi)
    print('mu:', mu)
    print('sigma:', sigma)

    # visualize
    #plt.figure(figsize=(12,8))
    for i in range(N):
        plt.scatter(data[i,0], data[i,1], s=30, c=gamma[i], alpha=0.5, marker="+")

    for i in range(K):
        ax = plt.axes()
        ax.arrow(mu_prev[i, 0], mu_prev[i, 1], mu[i, 0]-mu_prev[i, 0], mu[i, 1]-mu_prev[i, 1],
                  lw=0.8, head_width=0.02, head_length=0.02, fc='k', ec='k')
        plt.scatter([mu_prev[i, 0]], [mu_prev[i, 1]], c=c[i], marker='o', alpha=0.8)
        plt.scatter([mu[i, 0]], [mu[i, 1]], c=c[i], marker='o', edgecolors='k', linewidths=1)
        #plt.scatter([mu[i, 0]], [mu[i, 1]], c=c[i], marker='o')
    plt.title("step:{}".format(nframe))
    
    print_gmm_contour(mu, sigma, pi, K)
    # plt.show()
    
    if np.abs(diff) < 0.0001:
        plt.title('likelihood is converged.')
    else:
        plt.title("iter:{}".format(nframe-3))

def blog_demo():
    generate_demo_data()
    plt.show()

    initilize()
    plt.show()

    fig = plt.figure(figsize=(12,5))
    anim = ani.FuncAnimation(fig, animate, frames=48)
    # anim.save('gmm_anim.gif', writer='imagemagick', fps=3, dpi=128)
    plt.show()
